
# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from datetime import datetime
from datetime import timezone
from logging import getLogger
from logging import Logger

from neuro_san.interfaces.coded_tool import CodedTool


class DateTime(CodedTool):
    """
    CodedTool implementation which provides a current date and time
    """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.

                The argument dictionary expects the following keys:
                    None

        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.

                Keys expected for this implementation are:
                    None

        :return:
            In case of successful execution:
                The current UTC date and time as a string.
        """
        # Get current UTC time
        now: datetime = datetime.now(timezone.utc)

        # Format it in a user-friendly way
        friendly_time: str = now.strftime("%A, %B %d, %Y at %I:%M %p %Z")

        logger: Logger = getLogger(self.__class__.__name__)
        logger.debug("Current UTC date and time: %s", friendly_time)

        return str(now)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from leaf_common.config.config_filter_chain import ConfigFilterChain

from neuro_san.internals.graph.persistence.manifest_dict_config_filter import ManifestDictConfigFilter
from neuro_san.internals.graph.persistence.manifest_key_config_filter import ManifestKeyConfigFilter
from neuro_san.internals.graph.persistence.served_manifest_config_filter import ServedManifestConfigFilter


class ManifestFilterChain(ConfigFilterChain):
    """
    ConfigFilterChain for manifest files
    """

    def __init__(self, manifest_file: str):
        """
        Constructor

        :param manifest_file: The name of the manifest file we are processing for logging purposes
        """
        super().__init__()

        # Order matters
        self.register(ManifestKeyConfigFilter(manifest_file))
        self.register(ManifestDictConfigFilter(manifest_file))
        self.register(ServedManifestConfigFilter(manifest_file, warn_on_skip=True, entry_for_skipped=True))




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import argparse
import os

from neuro_san import REGISTRIES_DIR
from neuro_san.client.simple_one_shot import SimpleOneShot
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.test.assessor.assessor_assert_forwarder import AssessorAssertForwarder
from neuro_san.test.driver.data_driven_agent_test_driver import DataDrivenAgentTestDriver


class Assessor:
    """
    Command line tool for assessing an agent network's response
    to an input test case.
    """

    def __init__(self):
        """
        Constructor
        """
        self.args = None

    def main(self):
        """
        Main entry point for command line user interaction.
        """
        # Parse command line arguments
        self.parse_args()

        # Run the tests per the test case, collecting failure information in the AssertForwarder
        asserts = AssessorAssertForwarder()
        test_name: str = None
        test_name, _ = os.path.splitext(os.path.basename(self.args.test_hocon))
        driver = DataDrivenAgentTestDriver(asserts, test_name=test_name)

        print(f"Testing {self.args.test_hocon}:")
        driver.one_test(self.args.test_hocon)

        # Get raw failure information from AssertForwarder
        num_total: int = asserts.get_num_total()
        fail: List[Dict[str, Any]] = asserts.get_fail_dicts()
        num_pass: int = num_total - len(fail)

        # Initial output
        print(f"{num_pass}/{num_total} attempts passed.")
        if num_total == 1:
            print("There was only one test attempt done. Consider setting 'success_ratio' on your test case hocon.")
        if num_pass == num_total:
            return

        # Assess the failures by agent and output the results.
        assessment: Dict[str, List[Dict[str, Any]]] = self.assess_failures(fail)
        self.output_failures(assessment, num_total)

    def parse_args(self):
        """
        Parse command line arguments into member variables
        """
        arg_parser = argparse.ArgumentParser()
        self.add_args(arg_parser)
        self.args = arg_parser.parse_args()
        if self.args.assessor_agent is None:
            self.args.assessor_agent = REGISTRIES_DIR.get_file_in_basis("assess_failure.hocon")

    def add_args(self, arg_parser: argparse.ArgumentParser):
        """
        Adds arguments.  Allows subclasses a chance to add their own.
        :param arg_parser: The argparse.ArgumentParser to add.
        """
        arg_parser.add_argument("--test_hocon", type=str,
                                help="The test case .hocon file to use as a basis for assessment")
        arg_parser.add_argument("--assessor_agent", type=str, default=None,
                                help="""
The assessor agent to use. A default of None implies use of neuro-san stock assess_failure.hocon
""")
        arg_parser.add_argument("--connection", default="direct", type=str,
                                choices=["direct", "http", "https"],
                                help="""
The type of connection to initiate. Choices are to connect to:
    "http"      - an agent service via HTTP. Needs host and port.
    "https"     - an agent service via secure HTTP. Needs host and port.
    "direct"    - a session via library.
""")
        arg_parser.add_argument("--host", type=str, default=None,
                                help="hostname setting if not running locally")
        arg_parser.add_argument("--port", type=int, default=AgentSession.DEFAULT_PORT,
                                help="TCP/IP port to run the Agent service on")

    def categorize_one_failure(self, fail: Dict[str, Any], failure_modes: List[str]) -> str:
        """
        Categorize a single failure instance
        :param fail: A failure dictionary from asserts.get_fail_dicts()
        :param failure_modes: A List of strings describing known failure modes
        :return: A string describing an existing mode of failure or
                a description of a new mode of failure
        """

        text: str = f"""
The acceptance_criteria is:
"{fail.get('acceptance_criteria')}".

The text_sample is:
"{fail.get('text_sample')}".

The known failure_modes are:
{failure_modes}
"""

        # Use the "gist" agent to do the evalution
        one_shot = SimpleOneShot(self.args.assessor_agent, self.args.connection,
                                 host=self.args.host, port=self.args.port)
        raw_answer: str = one_shot.get_answer_for(text)
        return raw_answer

    def assess_failures(self, fail: List[Dict[str, Any]]) -> Dict[str, List[Dict[str, Any]]]:
        """
        Assess the failures in a failure list

        :param fail: A list of failure dictionaries from asserts.get_fail_dicts()
        :return: A dictionary whose keys describe specific modes of failure,
                and whose values are a list of failure dictionaries.
        """
        # Loop through each failure dictionary and categorize each one.
        assessment: Dict[str, Dict[str, Any]] = {}
        for one_failure in fail:

            # Always use our latest-greatest list of known failure modes.
            failure_modes: List[str] = assessment.keys()
            failure_mode: str = self.categorize_one_failure(one_failure, failure_modes)

            # If we have seen this failure before, add to the existing list.
            if failure_mode in failure_modes:
                assessment[failure_mode].append(one_failure)
            else:
                # ... otherwise make a new list for more potential matches later.
                assessment[failure_mode] = [one_failure]

        return assessment

    def output_failures(self, assessment: Dict[str, Dict[str, Any]], num_total: int):
        """
        Output the results of the failure assessment
        :param assessment: A dictionary whose keys describe specific modes of failure,
                and whose values are a list of failure dictionaries.
        :param num_total: The total number of attempts, including passing attempts.
        """
        # Get a total failure count
        num_fail: int = 0
        for failure_list in assessment.values():
            num_fail += len(failure_list)
        print(f"{num_fail}/{num_total} attempts failed.")

        # Sort the modes of failure by how often they occurred,
        # with the most common appearing at the beginning of the list.
        # The item[1] in the lambda is a list of failure dictionaries corresponding to the failure mode.
        # What is returned from sorted() is a list of key/value tuples which are reassembled
        # back into a sorted dictionary where earlier entries reflects higher occurrence.
        sorted_assessment: Dict[str, List[Dict[str, Any]]] = \
            dict(sorted(assessment.items(), key=lambda item: len(item[1]), reverse=True))

        # Output specific modes of failure with example agent return values.
        print("Modes of failure:")
        for failure_mode, failure_list in sorted_assessment.items():
            mode_count: int = len(failure_list)
            percent: float = 100.0 * mode_count / num_fail
            print("")
            print(f"{mode_count}/{num_fail} failures ({percent:.2f}%):")
            print("Failure Mode:")
            print(f"{failure_mode}")
            for index, fail_dict in enumerate(failure_list):
                print("")
                print(f"    Example {index+1}:")
                text_sample: str = fail_dict.get("text_sample")
                text_sample = text_sample.strip()
                print(f"    {text_sample}")

            acceptance_criteria: str = failure_list[0].get("acceptance_criteria")
            acceptance_criteria = acceptance_criteria.strip()
            print("")
            print("    Acceptance Criteria:")
            print(f"    {acceptance_criteria}")


if __name__ == '__main__':
    Assessor().main()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT


class ErrorFormatter:
    """
    Interface describing how error formatting can be called.

    Errors can come from various levels of the system - either as a result
    of the agents, or the agent framework itself.  In all cases, it benefits
    consumers of the agents' output to have consistent formatting of error
    messages regardless of source.
    """

    def format(self, agent_name: str, message: str, details: str = None) -> str:
        """
        Format an error message

        :param agent_name: A string describing the name of the agent experiencing
                the error.
        :param message: The specific message describing the error occurrence.
        :param details: An optional string describing further details of how/where the
                error occurred.  Think: traceback.
        :return: String encapsulation and/or filter of the error information
                presented in the arguments.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import copy
from logging import Logger
from logging import getLogger
import threading

from langchain_core.tools import BaseTool
from langchain_mcp_adapters.client import MultiServerMCPClient
from neuro_san.internals.run_context.langchain.mcp.mcp_servers_info_restorer import McpServersInfoRestorer


class LangChainMcpAdapter:
    """
    Adapter class to fetch tools from a Multi-Client Protocol (MCP) server and return them as
    LangChain-compatible tools. This class provides static methods for interacting with MCP servers.
    """

    _mcp_info_lock: threading.Lock = threading.Lock()
    _mcp_servers_info: Dict[str, Any] = None

    def __init__(self):
        """
        Constructor
        """
        self.client_allowed_tools: List[str] = []
        self.logger: Logger = getLogger(self.__class__.__name__)

    @staticmethod
    def _load_mcp_servers_info():
        """
        Loads MCP servers information from a configuration file if not already loaded.
        """
        with LangChainMcpAdapter._mcp_info_lock:
            if LangChainMcpAdapter._mcp_servers_info is None:
                LangChainMcpAdapter._mcp_servers_info = McpServersInfoRestorer().restore()
                if LangChainMcpAdapter._mcp_servers_info is None:
                    # Something went wrong reading the file.
                    # Prevent further attempts to load info.
                    LangChainMcpAdapter._mcp_servers_info = {}

    async def get_mcp_tools(
            self,
            server_url: str,
            allowed_tools: Optional[List[str]] = None,
            headers: Optional[Dict[str, Any]] = None
    ) -> List[BaseTool]:
        """
        Fetches tools from the given MCP server and returns them as a list of LangChain-compatible tools.

        :param server_url: URL of the MCP server, e.g. https://mcp.deepwiki.com/mcp or http://localhost:8000/mcp/
        :param allowed_tools: Optional list of tool names to filter from the server's available tools.
                              If None, all tools from the server will be returned.
        :param headers: Optional dictionary of HTTP headers to include in the MCP requests.

        :return: A list of LangChain BaseTool instances retrieved from the MCP server.
        """
        if self._mcp_servers_info is None:
            self._load_mcp_servers_info()

        mcp_tool_dict: Dict[str, Any] = {
            "url": server_url,
            "transport": "streamable_http",
        }
        # Try to look up authentication details first from the sly data then from the MCP servers info.
        headers_dict: Dict[str, Any] = headers or self._mcp_servers_info.get(server_url, {}).get("http_headers")
        if headers_dict:
            if isinstance(headers_dict, dict):
                # Use a copy to avoid modifying the original headers dictionary.
                mcp_tool_dict["headers"] = copy.copy(headers_dict)
            else:
                self.logger.error("MCP client headers for server %s must be a dictionary.",  server_url)

        client = MultiServerMCPClient(
            {"server": mcp_tool_dict}
        )

        # The get_tools() method returns a list of StructuredTool instances, which are subclasses of BaseTool.
        # Internally, it calls load_mcp_tools(), which uses an `async with create_session(...)` block.
        # This guarantees that any temporary MCP session created is properly closed when the block exits,
        # even if an error is raised during tool loading.
        # See: https://github.com/langchain-ai/langchain-mcp-adapters/blob/main/langchain_mcp_adapters/tools.py#L164
        # Optimization:
        #   It's possible we might want to cache these results somehow to minimize tool calls.
        mcp_tools: List[BaseTool] = await client.get_tools()

        # If allowed_tools is provided, filter the list to include only those tools.
        client_allowed_tools: List[str] = allowed_tools
        if client_allowed_tools is None:
            # Check if MCP server info has a "tools" field to use as allowed tools.
            client_allowed_tools = self._mcp_servers_info.get(server_url, {}).get("tools", [])
        # If client allowed tools is an empty list, do not filter the tools.
        if client_allowed_tools:
            mcp_tools = [tool for tool in mcp_tools if tool.name in client_allowed_tools]

        self.client_allowed_tools = client_allowed_tools

        for tool in mcp_tools:
            # Add "langchain_tool" tags so journal callback can idenitify it.
            # These MCP tools are treated as Langchain tools and can be reported in the thinking file.
            tool.tags = ["langchain_tool"]

        return mcp_tools




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from langchain_core.messages.base import BaseMessage


class TracedMessage(BaseMessage):
    """
    Absstract BaseMessage implementation of a message that is to be sent over
    to be displayed in an observability/tracing service like LangSmith.

    Note that these messages are intended to go to other services
    so any sensitive information should be redacted via an overridden
    translate_for_trace() method.

    As of 10/25/25:
    LangSmith will only show content in its trace viewer even when there are other
    fields filled in, even in additional kwargs. So this class makes all those fields
    visible by copying anything displayable into additional_kwargs and nulling out
    the content field when a trace_source message is provided in the constructor..
    """

    # Note this is an abstract class so we do not even define the type for it.

    def __init__(self, content: Union[str, List[Union[str, Dict]]] = "",
                 trace_source: TracedMessage = None,
                 **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        Args:
            content: The string contents of the message.
            trace_source: Another TracedMessage to copy additional_kwargs from.
            kwargs: Additional fields to pass to the
        """
        if trace_source:
            # If the content is set to something other than None or an empty string,
            # that is all that LangSmith will ever show.  So put what we want to show
            # in the additional_kwargs with effectively null content.
            additional_kwargs: Dict[str, Any] = trace_source.minimal_additional_kwargs()
            super().__init__(content="", additional_kwargs=additional_kwargs, **kwargs)
        else:
            super().__init__(content=content, **kwargs)

    @property
    def lc_serializable(self) -> bool:
        """
        Indicates if the object can be serialized by LangChain.
        """
        return True

    @property
    def lc_kwargs(self) -> Dict[str, Any]:
        """
        :return: the dictionary of keyword arguments for serialization.
        """
        raise NotImplementedError

    def minimal_additional_kwargs(self) -> Dict[str, Any]:
        """
        Creates a minimal additional_kwargs dictionary from the trace_source
        :return: The minimal kwargs dictionary
        """

        additional_kwargs: Dict[str, Any] = {}

        for key, value in self.lc_kwargs.items():
            new_key, new_value = self.translate_for_trace(key, value)
            if new_key is not None:
                additional_kwargs[new_key] = new_value

        return additional_kwargs

    def translate_for_trace(self, key: str, value: Any) -> Tuple[str, Any]:
        """
        :param key: The key to consider/translate.
        :param value: The value to consider/translate
        :return: A tuple with the new key and new value to be shown in the trace.
                New keys that are None are not included in the additional_kwargs.
                The default implementation simply ensures that there is something in the
                value to trace display to maximize information.
        """
        displayable: bool = value is not None and len(value) > 0
        if not displayable:
            return None, None

        return key, value




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from asyncio import Event
from json import dumps

from neuro_san.interfaces.reservation import Reservation
from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.reservations.agent_reservation import AgentReservation
from neuro_san.internals.validation.network.manifest_network_validator import ManifestNetworkValidator


class AccumulatingAgentReservationist(Reservationist):
    """
    A Reservationist implementation that accumulates Reservations to be deployed later.
    This *must* be used within the context of a Python async(hronous) event loop.

    This object supports the asynchronous Python Context Manager protocol by implementing the
    __aenter__() and __aexit__() methods.

    We expect its use within a CodedTool's async_invoke() to be:

        reservationist: Reservationist = args.get("reservationist")

        # Creating Reservations can be done outside the with-statement
        reservation: Reservation = await reservationist.reserve()
        deployments: Dict[Reservation, Dict[str, Any]] = {
            reservation: my_agent_spec,
            ...
        }

        # Deploy the reservations with confirmation event
        deployed_event: asyncio.Event = None
        async with reservationist:
            deployed_event = await reservationist.deploy(deployments, confirmation=True)

        if deployed_event is not None:
            await deployed_event.wait()
    """

    def __init__(self, supporting_reservationist: Reservationist,
                 agent_name: str,
                 max_lifetime_in_seconds: float = Reservationist.DEFAULT_LIFETIME):
        """
        Constructor

        :param supporting_reservationist: The supporting Reservationist that will do the real work
        :param agent_name: The full name of the agent using this instance
        :param max_lifetime_in_seconds: The maximum lifetime allowed for any Reservation.
        """
        self.max_lifetime_in_seconds: float = max_lifetime_in_seconds
        self.agent_name: str = agent_name
        self.deployments: Dict[Union[Event, str], Dict[Reservation, Dict[str, Any]]] = {}
        self._supporting_reservationist: Reservationist = supporting_reservationist
        self._supporting_reservation: Reservation = None
        self._external_networks: List[str] = None
        self._mcp_servers: List[str] = None

    async def reserve(self, lifetime_in_seconds: float = Reservationist.DEFAULT_LIFETIME,
                      prefix: str = "") -> Reservation:
        """
        Creates a reservation for a specific time period.

        :param lifetime_in_seconds: The lifetime of the reservation
        :param prefix: A string prefix to prepend to the id so as to provide external context.
        :return: A new Reservation object corresponding to the request.
        """
        actual_lifetime: float = min(lifetime_in_seconds, self.max_lifetime_in_seconds)
        if prefix is not None and len(prefix) > 0:
            self.validate_id(prefix)
        reservation = AgentReservation(actual_lifetime, prefix)

        return reservation

    def validate_with(self, external_networks: List[str] = None, mcp_servers: List[str] = None) -> Reservationist:
        """
        Allow passing of validation parameters when using a with-statement.
        These get reset in __aexit__().

        :param external_networks: A list of external network names
        :param mcp_servers: A list of MCP servers, as read in from a mcp_info.hocon file
        :return: The instance
        """
        self._external_networks = external_networks
        self._mcp_servers = mcp_servers
        return self

    async def __aenter__(self):
        """
        Python context manager protocol entrypoint.
        This is what gets called when you enter a with-statement.
        """
        # Clear any earlier deployments
        self.deployments = {}
        self._supporting_reservation = await self._supporting_reservationist.reserve(prefix=self.agent_name)

    async def deploy(self, deployment_dict: Dict[Reservation, Dict[str, Any]], confirmation: bool = False) -> Event:
        """
        Deploy multiple Reservations.

        :param deployment_dict: A dictionary whose keys are Reservation objects
                    previously obtained with reserve() and whose values are each an
                    object to deploy using the key as Reservation.
        :param confirmation: When True, indicates a confirmation Event is desired.
                            When False (the default), no such confirmation is needed.
        :return: Either a confiramtion Event object whose wait() method will confirm
                that the deployment is complete, or None, depending on the value passed
                for confirmation.
        """
        if not deployment_dict:
            return None

        # Do some validation here
        validator = ManifestNetworkValidator(external_network_names=self._external_networks,
                                             mcp_servers=self._mcp_servers)
        errors: Dict[str, List[str]] = {}
        for reservation, agent_network_spec in deployment_dict.items():

            # Validate the reservation id
            key: str = reservation.get_reservation_id()
            self.validate_id(key)

            # Validate what is being reserved.
            # Currently, we are assuming everything is an agent network
            new_errors: List[str] = validator.validate(agent_network_spec)
            if new_errors is not None and len(new_errors) > 0:
                # There were errors. Report all at once
                errors[key] = new_errors

        if errors:
            raise ValueError(f"Found {len(errors.keys())} validation errors when attempting to deploy():\n" +
                             f"{dumps(errors, indent=4, sort_keys=True)}")

        event: Event = None
        key: Union[str, Event] = None
        if confirmation:
            # Create an async.Event. This will be the key for the dictionary
            event = Event()
            key = event
        else:
            # Use the ID of the first Reservation as the key. This will be unique.
            first_res: Reservation = deployment_dict.keys()[0]
            key = first_res.get_reservation_id()
            # event is already None

        entry: Dict[Union[str, Event], Dict[Reservation, Dict[str, Any]]] = {
            key: deployment_dict
        }
        self.deployments.update(entry)

        return event

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Python context manager protocol exit point.
        This is what gets called when you exit a with-statement.
        This triggers the initiation of the deployment.
        """
        # Initiate the deployments
        await self._supporting_reservationist.deploy_one(self._supporting_reservation, self.deployments)

        # Reset state
        self._supporting_reservation = None
        self._external_networks = None
        self._mcp_servers = None
        self.deployments = {}

    @staticmethod
    def validate_id(test_id: str):
        """
        :param test_id:  The id string to test

        Will assert if the given id is not valid for external consumption
        """
        if test_id is None:
            raise ValueError("Reservation id is None")
        if len(test_id) == 0:
            raise ValueError("Reservation id is empty")

        invalid_chars: str = "/: "
        for invalid_char in invalid_chars:
            if invalid_char in test_id:
                # Probably more characters to look for w/rt url validity.
                raise ValueError(f"Reservation id {test_id} cannot contain any of these characters: '{invalid_chars}'")




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.interfaces.invocation_context import InvocationContext


class FrontMan:
    """
    Interface that describes how a chat interface can interact with a FrontMan
    """

    async def create_any_resources(self):
        """
        Creates resources that will be used throughout the lifetime of the component.
        """
        raise NotImplementedError

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        raise NotImplementedError

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        raise NotImplementedError

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        raise NotImplementedError

    async def submit_message(self, user_input: str) -> List[BaseMessage]:
        """
        Entry-point method for callers of the root of the Activation tree.

        :param user_input: An input string from the user.
        :return: A list of response messages for the run
        """
        raise NotImplementedError

    async def delete_any_resources(self):
        """
        Cleans up after any allocated resources
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

DEFAULT_HTTP_CONNECTIONS_BACKLOG: int = 128
DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS: int = 3600
DEFAULT_HTTP_SERVER_INSTANCES: int = 1
DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS: int = 0


class HttpServerConfig:
    """
    Class aggregating Tornado http server run-time configuration parameters.
    """

    def __init__(self):
        self.http_connections_backlog: int = DEFAULT_HTTP_CONNECTIONS_BACKLOG
        self.http_idle_connection_timeout_seconds: int = DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS
        self.http_server_instances: int = DEFAULT_HTTP_SERVER_INSTANCES
        self.http_port: int = 80
        self.http_server_monitor_interval_seconds: int = DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san.internals.persistence.abstract_async_config_restorer import AbstractAsyncConfigRestorer


class RawManifestRestorer(AbstractAsyncConfigRestorer):
    """
    Implementation of the AbstractAsyncConfigRestorer interface that reads the contents
    of a single manifest file for agent networks/registries.
    The restore() and async_restore() methods both return a dictionary.
    """

    def __init__(self):
        """
        Constructor
        """
        super().__init__(file_purpose="agent network manifest", env_var="AGENT_MANIFEST_FILE", must_exist=False)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Union

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector


class AllowUtil:
    """
    Utility class for handling questions about AgentNetwork allow blocks.
    """

    @staticmethod
    def is_allowed(inspector: AgentNetworkInspector, allow_key: str, deeper_keys: List[str] = None) -> bool:
        """
        Check if the agent network allows the given key on any one of its agent nodes.

        :param inspector: The AgentNetworkInspector instance.  Common internal implementations of
                        AgentNetworkInspector include AgentNetwork, AgentToolRegistry.
        :param allow_key: The key to check in the allow dictionary.
        :param deeper_keys: A list of keys to look deeper in the agent spec.

        :return: True if the key is present and true somewhere in the agent network.
                 False otherwise.
        """
        allow: bool = False

        if inspector is None or allow_key is None:
            return allow

        # Get the list of tools to look for
        config: Dict[str, Any] = inspector.get_config()
        tools: List[Dict[str, Any]] = []
        tools = config.get("tools", tools)

        # Loop through each of the tools/agents
        empty: Dict[str, Any] = {}
        agent_spec: Dict[str, Any] = empty
        for agent_spec in tools:
            allow = AllowUtil.is_allowed_in_agent_spec(agent_spec, allow_key, deeper_keys)
            if allow:
                return True

        return allow

    @staticmethod
    def is_allowed_in_agent_spec(agent_spec: Dict[str, Any], allow_key: str, deeper_keys: List[str] = None) -> bool:
        """
        Check if the agent spec allows the given key.
        Will also search each of the deeper_keys and can treat each value as a list to inspect as well
        (ala middleware).

        :param agent_spec: The agent spec to check.
        :param allow_key: The key to check in the allow dictionary.
        :param deeper_keys: A list of keys to look deeper in the agent spec.

        :return: True if the key is present and true somewhere in the agent network.
                 False otherwise.
        """
        if agent_spec is None or allow_key is None:
            return False

        empty: Dict[str, Any] = {}

        # Compile a key to look for
        look_for_key: str = allow_key
        if not look_for_key.startswith("allow."):
            look_for_key = f"allow.{allow_key}"

        # Look in the agent spec for the allow key
        allow_extractor = DictionaryExtractor(agent_spec)
        if allow_extractor.get(look_for_key, False):
            return True

        # Any deeper keys? Return early if not.
        if deeper_keys is None or len(deeper_keys) == 0:
            return False

        # Maybe look in the deeper_keys for the allow key
        for deeper_key in deeper_keys:

            # Make the value a list if we don't get one to consolidate on iterative logic
            deeper_list: Union[List[Dict[str, Any]], Dict[str, Any]] = agent_spec.get(deeper_key, empty)
            if not isinstance(deeper_list, list):
                deeper_list = [deeper_list]

            for deeper_spec in deeper_list:
                deeper_extractor = DictionaryExtractor(deeper_spec)
                if deeper_extractor.get(look_for_key, False):
                    return True

        return False




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.messages.chat_message_type import ChatMessageType


class MessageProcessor:
    """
    An interface for processing a single message.
    """

    def reset(self):
        """
        Resets any previously accumulated state
        """

    def should_block_downstream_processing(self, chat_message_dict: Dict[str, Any],
                                           message_type: ChatMessageType) -> bool:
        """
        :param chat_message_dict: The ChatMessage dictionary to consider.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        :return: True if the given message should be blocked from further downstream
                processing.  False otherwise (the default).
        """
        _ = chat_message_dict, message_type
        return False

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        raise NotImplementedError

    async def async_process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message asynchronously.
        By default, this simply calls the synchronous version.

        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        self.process_message(chat_message_dict, message_type)

    def process_messages(self, chat_message_dicts: List[Dict[str, Any]]):
        """
        Convenience method for processing lists of messages.
        :param chat_message_dicts: The messages to process.
        """
        for message in chat_message_dicts:
            message_type: ChatMessageType = message.get("type")
            self.process_message(message, message_type)

    async def async_process_messages(self, chat_message_dicts: List[Dict[str, Any]]):
        """
        Convenience method for asynchronouslt processing lists of messages.
        :param chat_message_dicts: The messages to process.
        """
        for message in chat_message_dicts:
            message_type: ChatMessageType = message.get("type")
            await self.async_process_message(message, message_type)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san import TOP_LEVEL_DIR
from neuro_san.internals.persistence.abstract_async_config_restorer import AbstractAsyncConfigRestorer


class ToolboxInfoRestorer(AbstractAsyncConfigRestorer):
    """
    Implementation of the AbstractAsyncConfigRestorer interface to read in a ToolboxInfo dictionary
    instance given a hocon file name.
    Both restore() and async_restore() return dictionaries.
    """

    def __init__(self):
        """
        Constructor
        """
        super().__init__(file_purpose="toolbox_info")

    def get_file_path(self, file_reference: str = None) -> str:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: a string representing the file path to use
        """
        use_file: str = file_reference

        if file_reference is None or len(file_reference) == 0:
            # Read from the default
            use_file = TOP_LEVEL_DIR.get_file_in_basis("internals/run_context/langchain/toolbox/toolbox_info.hocon")

        return use_file





# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from leaf_common.config.file_of_class import FileOfClass

# Normally we don't use __init__.py files to define anything,
# but here we define some constants that point to important directories in the distribution.
TOP_LEVEL_DIR = FileOfClass(__file__)
DEPLOY_DIR = FileOfClass(__file__, path_to_basis="./deploy")
REGISTRIES_DIR = FileOfClass(__file__, path_to_basis="./registries")




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from grpc import StatusCode
from grpc.aio import AioRpcError

from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext


class ExternalToolAdapter:
    """
    Class handles setting up a connection to an external agent server
    so that its agents can be used as tools.
    """

    def __init__(self, session_factory: AsyncAgentSessionFactory, agent_url: str):
        """
        Constructor

        :param agent_url: The URL describing where to find the desired agent.
        """

        self.agent_url: str = agent_url
        self.session_factory: AsyncAgentSessionFactory = session_factory
        self.function_json: Dict[str, Any] = None

    async def get_function_json(self, invocation_context: InvocationContext) -> Dict[str, Any]:
        """
        :param invocation_context: The context policy container that pertains to the invocation
        :return: The function json for the agent, as specified by the external agent.
        """
        if self.function_json is None:

            # Lazily get the information about the service
            session: AgentSession = self.session_factory.create_session(self.agent_url,
                                                                        invocation_context=invocation_context)

            # Set up the request. Turns out we don't need much.
            request_dict: Dict[str, Any] = {}

            # Get the function spec so we can call it as a tool later.
            try:
                function_response: Dict[str, Any] = await session.function(request_dict)
                self.function_json = function_response.get("function")
            except (AioRpcError, ValueError) as exception:
                message: str = f"Problem accessing external agent {self.agent_url}.\n"
                if not isinstance(exception, AioRpcError) or exception.code() == StatusCode.UNIMPLEMENTED:
                    message += """
The server (which could be your own localhost) is currently not serving up
an agent network by that name. Try these hints:
1. Check to see that you do not have a typo in your reference to the external agent
   in the calling hocon file.
2. If you have control over the server, check to see that the agent network your are trying
   to reach has an entry in the manifest.hocon file whose value is set to true.
3. Consider restarting the server, as perhaps a server does not continually look
   for changes to hocon files or manifest files during normal operation.
4. If you are new to calling external agent networks, know that:
    a. Simply referencing an agent within your own hocon file with a / prefix
       does not mean the server is serving that agent up separately.
    b. Every agent that is externally referenceable needs its own hocon file
       which must also have an entry in the manifest.hocon file for the server.
    c. There is one and only one "front man" agent note in each network described
       by a hocon file that receives input on behalf of the network.
    d. In order to be called by external agents, that front man must have a full
       "function" definition, which includes a description, and at least one parameter
       defined.  These are how calling agents know how to interact with the agent network.
"""
                raise ValueError(message) from exception

        return self.function_json




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from neuro_san.internals.messages.traced_message import TracedMessage


class AgentMessage(TracedMessage):
    """
    TracedMessage implementation of a message from an agent
    """

    structure: Optional[Dict[str, Any]] = None

    type: Literal["agent"] = "agent"

    def __init__(self, content: Union[str, List[Union[str, Dict]]] = "",
                 structure: Dict[str, Any] = None,
                 trace_source: AgentMessage = None,
                 **kwargs: Any) -> None:
        """
        Constructor

        :param content: The string contents of the message.
        :param structure: A dictionary to pack into the message
        :param trace_source: A message of the same type to prepare for tracing display
        :param kwargs: Additional fields to pass to initialize
        """
        super().__init__(content=content, trace_source=trace_source, **kwargs)
        self.structure: Dict[str, Any] = structure

    @property
    def lc_kwargs(self) -> Dict[str, Any]:
        """
        :return: the keyword arguments for serialization.
        """
        return {
            "content": self.content,
            "structure": self.structure,
        }




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import uuid

from aiohttp.client_exceptions import ClientConnectionError

from langchain_core.messages.base import BaseMessage

from leaf_common.parsers.field_extractor import FieldExtractor

from neuro_san.internals.graph.activations.argument_assigner import ArgumentAssigner
from neuro_san.internals.graph.activations.calling_activation import CallingActivation
from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.graph.interfaces.callable_activation import CallableActivation
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class BranchActivation(CallingActivation, CallableActivation):
    """
    A CallingActivation subclass which can also be a CallableActivation.
    Thus, instances are able to be branch nodes in the tool call graph.
    Leaf nodes in the call graph are also these guys, they just happen to
    not call anyone else.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, parent_run_context: RunContext,
                 factory: AgentToolFactory,
                 arguments: Dict[str, Any],
                 agent_tool_spec: Dict[str, Any],
                 sly_data: Dict[str, Any]):
        """
        Constructor

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param factory: The AgentToolFactory used to create tools
        :param arguments: A dictionary of the tool function arguments passed in
        :param agent_tool_spec: The dictionary describing the JSON agent tool
                            to be used by the instance
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        """
        super().__init__(parent_run_context, factory, agent_tool_spec, sly_data)
        self.arguments: Dict[str, Any] = arguments

    def get_assignments(self) -> str:
        """
        :return: The string prompt for assigning values to the arguments to the agent.
        """
        # Get the properties of the function
        extractor: FieldExtractor = FieldExtractor()
        empty: Dict[str, Any] = {}

        agent_spec: Dict[str, Any] = self.get_agent_tool_spec()

        # Properties describe the function arguments
        properties: Dict[str, Any] = extractor.get_field(agent_spec, "function.parameters.properties", empty)

        assigner = ArgumentAssigner(properties)
        # The assigner will skip any arguments that the value is None.
        assignments: List[str] = assigner.assign(self.arguments)

        # Start to build a single assignments string, with one sentence for each property
        # listed (exception for name and description).
        assignments_str: str = "\n".join(assignments)
        return assignments_str

    def get_command(self) -> str:
        """
        :return: A string describing the objective of the component.
        """
        agent_spec: Dict[str, Any] = self.get_agent_tool_spec()

        # The command will be combined with assignments (arguments from an upstream agent)
        # to direct the agent toward a specific task.
        # Optional: if omitted, the agent will choose an action based on the assignments
        # and the list of available tools.
        return agent_spec.get("command")

    async def integrate_callable_response(self, run: Run, messages: List[BaseMessage]) -> List[BaseMessage]:
        """
        :param run: The Run for the prescriptor (if any)
        :param messages: A current list of messages for the component.
        :return: An updated list of messages after this operation is done.
                This default implementation is just a pass-through of the messages argument.
        """
        new_messages: List[BaseMessage] = messages

        callable_tool_names: List[str] = self.get_callable_tool_names(self.agent_tool_spec)
        if callable_tool_names is None:
            # If there are no callable_tool_names, then there is no action from the
            # callable class to integrate
            return new_messages

        while run.requires_action():
            # The tool we just called requires more information
            new_run: Run = await self.make_tool_function_calls(run)
            new_run = await self.run_context.wait_on_run(new_run, self.journal)
            new_messages = await self.run_context.get_response()

        return new_messages

    async def build(self) -> BaseMessage:
        """
        Main entry point to the class.

        :return: A BaseMessage produced during this process.
        """

        assignments: str = self.get_assignments()
        instructions: str = self.get_instructions()

        uuid_str: str = str(uuid.uuid4())
        component_name: str = self.get_name()
        unique_name: str = f"{uuid_str}_{component_name}"
        await self.create_resources(unique_name, instructions, None)

        # If there is command, combine it with assignment to be used as HumanMessage.
        command: str = self.get_command()
        if command:
            assignments = assignments + "\n" + command

        run: Run = await self.run_context.submit_message(assignments)
        run = await self.run_context.wait_on_run(run, self.journal)

        messages: List[BaseMessage] = await self.run_context.get_response()

        messages = await self.integrate_callable_response(run, messages)

        # Return the last message
        return messages[-1]

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return self.run_context.get_origin()

    async def use_tool(self, tool_name: str, tool_args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        Experimental method to call a tool more directly from a subclass.

        NOTE: This method is not correctly reporting history or anything like that just yet.

        :param tool_name: The name of the tool to invoke
        :param tool_args: A dictionary of arguments to send to the tool.
        :param sly_data: private data dictionary to send to the tool.
        :return: A string representing the last received content text of the last message.
        """

        # Use the tool
        our_agent_spec: Dict[str, Any] = self.get_agent_tool_spec()
        callable_activation: CallableActivation = self.factory.create_agent_activation(self.run_context,
                                                                                       our_agent_spec,
                                                                                       tool_name,
                                                                                       sly_data,
                                                                                       tool_args)
        message: BaseMessage = None
        try:
            # DEF - need to integrate sly_data
            message = await callable_activation.build()

        except ClientConnectionError as exception:
            # There is a case where we could give a little more help.
            invocation_context: InvocationContext = self.run_context.get_invocation_context()
            async_factory: AsyncAgentSessionFactory = invocation_context.get_async_session_factory()
            if not async_factory.is_use_direct() and tool_name.startswith("/"):
                # Special case where we can give a hint about using direct
                raise ValueError(f"""
Attempt to call {tool_name} as an external agent network over http failed.
If you are getting this from the agent_cli command line, consider adding the --local_externals_direct
flag to your invocation.
""") from exception

            # Nope. Just a regular http connection failure given the tool_name. Can't help ya.
            raise exception

        # We got a message back, take the content as the return string
        return message.content

    async def use_reservation(self, reservation_id: str, args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        Experimental method to call a reserved temporary network more directly from a subclass.

        NOTE: This method is not correctly reporting history or anything like that just yet.

        :param reservation_id: The string id of the reservation to invoke
        :param args: A dictionary of arguments to send to the reserved temporary network.
        :param sly_data: private data dictionary to send to the reserved temporary network.
        :return: A string representing the last received content text of the last message.
        """
        # DEF - for now just assume a local copy
        result: str = await self.use_tool(f"/{reservation_id}", args, sly_data)
        return result




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Dict
from typing import List

# Dictionary with provider key env var -> strings to look for
API_KEY_EXCEPTIONS: Dict[str, List] = {
    "OPENAI_API_KEY": ["OPENAI_API_KEY", "Incorrect API key provided"],
    "ANTHROPIC_API_KEY": ["ANTHROPIC_API_KEY", "anthropic_api_key", "invalid x-api-key", "credit balance"],
    "GOOGLE_API_KEY": ["Application Default Credentials", "default credentials", "Gemini: 400 API key not valid"],

    # Azure OpenAI requires several parameters; all can be set via environment variables
    # except "deployment_name", which must be provided explicitly.
    "AZURE_OPENAI_API_KEY": ["invalid subscription key", "wrong API endpoint"],
    "AZURE_OPENAI_ENDPOINT": ["base_url", "azure_endpoint", "AZURE_OPENAI_ENDPOINT"],
    "OPENAI_API_VERSION": ["api_version", "OPENAI_API_VERSION"],
    "AZURE_OPENAI_DEPLOYMENT_NAME": ["API deployment for this resource does not exist"],
}

AZURE_DOCUMENTATION: str = "https://learn.microsoft.com/en-us/azure/ai-services/openai/"
"chatgpt-quickstart?tabs=keyless%2Ctypescript-keyless%2Cpython-new%2Ccommand-line&pivots=programming-language-python"

# Dictionary with provider key env var -> link to documentation
API_KEY_DOCUMENTATION: Dict[str, List] = {
    "AZURE_OPENAI_API_KEY": AZURE_DOCUMENTATION,
    "AZURE_OPENAI_ENDPOINT": AZURE_DOCUMENTATION,
    "OPENAI_API_VERSION": AZURE_DOCUMENTATION,
    "AZURE_OPENAI_DEPLOYMENT_NAME": AZURE_DOCUMENTATION,
}

INTERNAL_ERRORS_LIST: List[str] = ["bound to a different event loop"]


class ApiKeyErrorCheck:
    """
    Class for common policy when checking for API key errors for various LLM providers.
    """

    @staticmethod
    def check_for_api_key_exception(exception: Exception) -> str:
        """
        :param exception: An exception to check
        :return: A more helpful exception message if it relates to an API key or None
                if it does not pertain to an API key.
        """

        exception_message: str = str(exception)
        matched_keys: List[str] = []
        matched_documentation_link: str = ""

        # Collect all keys that have any associated string in the exception message
        # since there could be multiple keys with the exact same message.
        for api_key, string_list in API_KEY_EXCEPTIONS.items():
            for find_string in string_list:
                if find_string in exception_message:
                    matched_keys.append(api_key)
                    matched_documentation_link = API_KEY_DOCUMENTATION.get(api_key, "")
                    # No need to check the remaining strings for this key
                    break

        if matched_keys:
            keys_str = ", ".join(matched_keys)
            return f"""
A value for the {keys_str} environment variable must be correctly set in the neuro-san
server or run-time environment in order to use this agent network.

Some things to try:
1) Double check that your value for {keys_str} is set correctly
2) If you do not have a value for {keys_str}, visit the LLM provider's website to get one {matched_documentation_link}
3) It's possible that your credit balance on your account with the LLM provider is too low
   to make the request.  Check that.
4) Sometimes these errors happen because of firewall blockages to the site that hosts the LLM.
   Try checking that you can reach the regular UI for the LLM from a web browser
   on the same machine making this request.
"""

        return None

    @staticmethod
    def check_for_internal_error(exception_traceback: str) -> bool:
        """
        Check if exception traceback points to some internal LLM stack problem,
        not necessarily related to API keys being absent or invalid.
        This function used as an additional check while using check_for_api_key_exception()
        :param exception_traceback: exception traceback string;
        :return: True if exception seems to be caused by some internal problems,
                 False otherwise
        """
        for err_msg in INTERNAL_ERRORS_LIST:
            if err_msg in exception_traceback:
                return True
        return False



# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from json import JSONDecodeError
from json import loads
from logging import getLogger
from logging import Logger
from typing import Any
from typing import Dict
from typing_extensions import override

from langchain.agents.middleware import AgentMiddleware
from langchain.agents.middleware import AgentState
from langchain_core.messages import AIMessage

from neuro_san import REGISTRIES_DIR
from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.reservations.reservation_util import ReservationUtil


class NetworkCopyMiddleware(AgentMiddleware):
    """
    Middleware that copies an existing agent network into a temporary reservation.

    After the agent responds, this middleware parses the model's output as a JSON
    object with an "agent_name" key, loads the corresponding .hocon network spec,
    and creates a time-limited reservation for it via the Reservationist interface.
    The reservation details are stored in sly_data for downstream consumers.

    Note: The copied network is not currently invoked from within the middleware.
    Doing so is possible but would require passing the invocation context to the
    middleware and performing a dynamic tool call. This may be added in a future
    iteration. For now, this serves as a proof of concept for using the Reservations
    infrastructure in middleware.
    """

    def __init__(self, reservationist: Reservationist, sly_data: Dict[str, Any]) -> None:
        """
        Initialize the middleware.
        :param reservationist: Reservationist implementation that allows procurement of agent ids
                for a specific amount of time for direct sessions only.
        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.
        """
        self.reservationist = reservationist
        self.sly_data = sly_data
        self.logger: Logger = getLogger(self.__class__.__name__)

    @override
    async def aafter_agent(
        self,
        state: AgentState,
        runtime: Any
    ) -> Dict[str, Any] | None:
        """
        Copy the agent network provided by the model's response.

        :param state: Current agent state
        :param runtime: Runtime context
        :return: Dict with error message and jump directive, or None if valid
        """
        response: str = state.get("messages")[-1].content

        agent_name: str = self._parse_agent_name(response)
        if agent_name is None:
            return self._agent_response("Please provide the name of the network to copy.")

        network: AgentNetwork = self._restore_network(agent_name)
        if not network:
            self.logger.error("Cannot find agent %s", agent_name)
            return self._agent_response(
                f"Cannot find agent {agent_name}. "
                "Please provide a valid name of the network to copy."
            )

        lifetime_in_seconds: float = 5 * 60.0
        self.logger.info("Creating temporary network from %s", agent_name)
        reservation, error = await ReservationUtil.wait_for_one(
            {"reservationist": self.reservationist},
            network.get_config(),
            lifetime_in_seconds,
            prefix=f"copy_cat-{agent_name}"
        )

        if error is not None:
            self.logger.error("Error: %s", error)
            return self._agent_response(f"Got the following error: {error}.")

        reservation_id: str = reservation.get_reservation_id()
        lifetime_in_seconds = reservation.get_lifetime_in_seconds()
        self.logger.info("Successfully created temporary network %s", reservation_id)

        self.sly_data["agent_reservations"] = [
            {
                "reservation_id": reservation_id,
                "lifetime_in_seconds": lifetime_in_seconds,
                "expiration_time_in_seconds": reservation.get_expiration_time_in_seconds()
            }
        ]

        return self._agent_response(
            f"The temporary agent name is {reservation_id}. "
            f"Hurry, it's only available for {lifetime_in_seconds} seconds."
        )

    def _agent_response(self, message: str) -> Dict[str, Any]:
        """Return an agent response dict with the given message."""
        return {"messages": [AIMessage(message)]}

    def _parse_agent_name(self, response: str) -> str | None:
        """
        Parse the model response JSON and extract the agent name.

        :param response: Raw model response string
        :return: The agent name, or None if parsing/validation fails
        """
        try:
            args: Dict[str, str] = loads(response)
        except JSONDecodeError as json_error:
            self.logger.error("Cannot parse '%s' into JSON format. Got %s", response, json_error)
            return None

        agent_name: str = args.get("agent_name")
        if not agent_name or not isinstance(agent_name, str):
            self.logger.error("Need a non-empty string for agent name")
            return None

        if agent_name.endswith(".hocon"):
            agent_name = agent_name[:-6]

        return agent_name

    def _restore_network(self, agent_name: str) -> AgentNetwork | None:
        """
        Restore an agent network from the registry by name.

        :param agent_name: Agent name (with or without .hocon suffix)
        :return: The restored AgentNetwork, or None if not found
        """
        hocon_name: str = f"{agent_name}.hocon"
        copy_file: str = REGISTRIES_DIR.get_file_in_basis(hocon_name)
        restorer = AgentNetworkRestorer()
        return restorer.restore(file_reference=copy_file)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Type

from os import environ
from threading import Lock

from leaf_common.config.resolver_util import ResolverUtil

# Lazy loading of optional openfga_sdk types.
OpenFgaClient: Type[Any] = ResolverUtil.create_type("openfga_sdk.client.client.OpenFgaClient",
                                                    raise_if_not_found=False,
                                                    install_if_missing="openfga-sdk")


class OpenFgaStoreCache:
    """
    Per the Open FGA docs here:
        https://github.com/openfga/python-sdk?tab=readme-ov-file#initializing-the-api-client

        "We strongly recommend you initialize the OpenFgaClient only once
         and then re-use it throughout your app..."

    Singleton factory for providing client access to OpenFGA server.
    Clients are preserved on a per-thread + per-store basis.

    In production code there will really only ever be a single fact data store
    (DEFAULT_STORE_NAME) for all the threads, but we will want all the threads
    to have their own client.

    In testing, however, we have the opposite: multiple stores for (likely)
    single threaded tests so that they specifically do not stomp on anything
    real code would use.
    """

    # A mapping of store names (like we get in the env var)
    # to store ids (like we get back from the OpenFGA server to initialize clients with).
    store_name_to_id: Dict[str, str] = {}

    # Threaded lock - on purpose even though async access is used
    lock = Lock()

    # Store name to use when none is specified by the caller.
    DEFAULT_STORE_NAME: str = environ.get("FGA_STORE_NAME", "default")

    @classmethod
    async def get_client(cls, store_name: str = None) -> OpenFgaClient:
        """
        :param store_name: The store name to use for fact storage.
                We expect workaday client code to not pass this in, but we allow
                a different store name as an arg so that test code can talk to an
                existing server without messing anything real up.
        :return: a connection to the OpenFGA authorization server for the given store name/
                thread id combination..
        """
        if store_name is None:
            # This allows workaday code to not worry about store names,
            # including when it is called by unit tests.
            store_name = OpenFgaStoreCache.DEFAULT_STORE_NAME

        store_id: str = OpenFgaStoreCache.store_name_to_id.get(store_name)

        # Lazy loading of OpenFgaInit class which directly uses OpenFga SDK types.
        # pylint: disable=import-outside-toplevel
        from neuro_san.internals.authorization.openfga.open_fga_init import OpenFgaInit

        if store_id is None:
            # Note: Synchronous lock is required here
            init = OpenFgaInit()
            with OpenFgaStoreCache.lock:
                store_id = await init.initialize_store(store_name)
                OpenFgaStoreCache.store_name_to_id[store_name] = store_id

        fga_client: OpenFgaClient = OpenFgaInit.initialize_one_client(store_id=store_id)

        return fga_client

    @classmethod
    def _remove_key_for_testing(cls, store_name: str):
        """
        Removes a store_name key for testing purposes only.

        :param store_name: The store name to use for fact storage.
        :param remove_from_map: When True the entry will be
                removed from the store-name/thread map.
        """

        # Do not hold the lock as the caller will be holding for us.
        store_id: str = OpenFgaStoreCache.store_name_to_id.get(store_name)
        if store_id is not None:

            # Do not actually remove the default store as that is what the app
            # will be using. Remove any other store for testing though.
            if OpenFgaStoreCache.DEFAULT_STORE_NAME != store_name:
                del OpenFgaStoreCache.store_name_to_id[store_name]

    @classmethod
    def reset_for_testing(cls):
        """
        Reset the instance for testing purposes only.
        """
        with OpenFgaStoreCache.lock:

            if len(OpenFgaStoreCache.store_name_to_id) > 0:

                # Close all the clients registered in the map
                # Need to add remove_from_map=False or else will get this error:
                #        "RuntimeError: dictionary changed size during iteration"
                # pylint: disable=consider-iterating-dictionary
                for store_name in cls.store_name_to_id.keys():
                    cls._remove_store_name_for_testing(store_name)

                # Clear the map separately
                cls.store_name_to_id = {}




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from logging import getLogger
from logging import Logger

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator


class UrlNetworkValidator(AbstractNetworkValidator):
    """
    AbstractNetworkValidator that looks for correct URLs in an agent network
    """

    def __init__(self, external_agents: List[str] = None, mcp_servers: List[str] = None):
        """
        Constructor

        :param external_agents: A list of valid /external_agent referencess
        :param mcp_servers: A list of MCP servers, as read in from a mcp_info.hocon file
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.external_agents: List[str] = external_agents
        self.mcp_servers: List[str] = mcp_servers

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.
        Check if URL of MCP servers and external_agents are valid.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: List of errors indicating invalid URL
        """
        errors: List[str] = []

        # Compile list of urls to check
        urls: List[str] = []
        if self.external_agents:
            urls.extend(self.external_agents)
        if self.mcp_servers:
            urls.extend(self.mcp_servers)

        self.logger.info("Validating URLs for MCP tools and subnetwork...")

        for agent_name, agent in name_to_spec.items():
            if agent.get("tools"):
                tools: List[str] = agent.get("tools")
                if tools:
                    safe_tools: List[str] = self.remove_dictionary_tools(tools)
                    self.check_safe_urls(agent_name, safe_tools, urls, errors)

        return errors

    def check_safe_urls(self, agent_name: str, safe_tools: List[str], urls: List[str], errors: List[str]):
        """
        Check that urls are valid

        :param agent_name: Name of agent
        :param safe_tools: List of tools
        :param urls: List of URLs
        :param errors: List of errors. Potentially modified on exit.
        """
        for tool in safe_tools:
            # pylint: disable=too-many-boolean-expressions
            if self.is_url_or_path(tool) and \
                    tool not in urls and \
                    not tool.startswith("http://") and \
                    not tool.startswith("https://") and \
                    not tool.endswith("mcp") and \
                    not tool.endswith("mcp/"):
                error_msg = f"Agent '{agent_name}' has invalid URL or path in tools." + \
                            f" Invalid tool: '{tool}' urls: {urls}"
                errors.append(error_msg)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Type

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.anthropic_llm_policy import AnthropicLlmPolicy
from neuro_san.internals.run_context.langchain.llms.azure_llm_policy import AzureLlmPolicy
from neuro_san.internals.run_context.langchain.llms.bedrock_llm_policy import BedrockLlmPolicy
from neuro_san.internals.run_context.langchain.llms.gemini_llm_policy import GeminiLlmPolicy
from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy
from neuro_san.internals.run_context.langchain.llms.langchain_llm_factory import LangChainLlmFactory
from neuro_san.internals.run_context.langchain.llms.langchain_llm_resources import LangChainLlmResources
from neuro_san.internals.run_context.langchain.llms.nvidia_llm_policy import NvidiaLlmPolicy
from neuro_san.internals.run_context.langchain.llms.ollama_llm_policy import OllamaLlmPolicy
from neuro_san.internals.run_context.langchain.llms.openai_llm_policy import OpenAILlmPolicy


class StandardLangChainLlmFactory(LangChainLlmFactory):
    """
    Factory class for LLM operations

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default, this comes from
                                    the model description in this class.
    """

    def __init__(self, class_to_llm_policy_type: Dict[str, Type[LlmPolicy]] = None):
        """
        Constructor

        :param class_to_llm_policy_type: A dictionary mapping llm class names
                    used in an llm_info.hocon file to python types that represent
                    the correspoinding LlmPolicy implementation.

                    The default value is None, allowing for the stock lookup table
                    for the base library to be used for the instance.

                    Subclasses can pass in their own lookup table if desired,
                    making the extension of the library easier.
        """
        self.class_to_llm_policy_type: Dict[str, LlmPolicy] = class_to_llm_policy_type
        if self.class_to_llm_policy_type is None:
            self.class_to_llm_policy_type = {
                "anthropic": AnthropicLlmPolicy,
                "azure-openai": AzureLlmPolicy,
                "bedrock": BedrockLlmPolicy,
                "gemini": GeminiLlmPolicy,
                "nvidia": NvidiaLlmPolicy,
                "openai": OpenAILlmPolicy,
                "ollama": OllamaLlmPolicy,
            }

    def create_base_chat_model(self, config: Dict[str, Any]) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel from the fully-specified llm config.

        This method is provided for backwards compatibility.
        Prefer create_llm_resources() instead,
        as this allows server infrastructure to better account for outstanding
        connections to LLM providers when connections drop.

        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        raise NotImplementedError

    # pylint: disable=too-many-branches
    def create_llm_resources(self, config: Dict[str, Any]) -> LangChainLlmResources:
        """
        Create a BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :return: A LangChainLlmResources instance containing
                a BaseLanguageModel (can be Chat or LLM) and all related resources
                necessary for managing the model run-time lifecycle.
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        # pylint: disable=too-many-locals
        # Construct the LLM
        llm: BaseLanguageModel = None

        chat_class: str = config.get("class")
        if chat_class is not None:
            chat_class = chat_class.lower()

        # Check for key "model_name", "model", and "model_id" to use as model name
        # If the config is from default_llm_info, this is always "model_name"
        # but with user-specified config, it is possible to have the other keys will be specifed instead.
        model_name: str = config.get("model_name") or config.get("model") or config.get("model_id")

        # Get from table of policy classes
        llm_policy: LlmPolicy = None
        policy_class: Type[LlmPolicy] = self.class_to_llm_policy_type.get(chat_class)
        if policy_class is not None:

            # Use the LlmPolicy type we found in the lookup table to create an call a new instance.
            llm_policy = policy_class()
            llm, llm_policy = llm_policy.create_llm_resources_components(config)

        elif chat_class is None:
            raise ValueError(f"Class name {chat_class} for model_name {model_name} is unspecified.")
        else:
            raise ValueError(f"Class {chat_class} for model_name {model_name} is unrecognized.")

        # Return the LlmResources with the llm_policy that was created.
        # That might be None, and that's OK.
        return LangChainLlmResources(llm, llm_policy=llm_policy)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.interfaces.context_type_tracing_context_factory import ContextTypeTracingContextFactory
from neuro_san.internals.run_context.langchain.tracing.langchain_tracing_context_factory \
    import LangChainTracingContextFactory


class MasterTracingContextFactory:
    """
    Creates the correct kind of ContextTypeTracingContextFactory
    given the underlying toolkit.
    """

    @staticmethod
    def create_tracing_context_factory(config: Dict[str, Any] = None) -> ContextTypeTracingContextFactory:
        """
        Creates an appropriate ContextTypeTracingContextFactory

        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        :param run_target: The RunTarget instance to be traced
        :return: A ContextTypeTracingContextFactory appropriate for the context_type in the config.
        """

        tracing_context_factory: ContextTypeTracingContextFactory = None
        context_type: str = MasterTracingContextFactory.get_context_type(config)

        if context_type.startswith("openai"):
            tracing_context_factory = None
        elif context_type.startswith("langchain"):
            tracing_context_factory = LangChainTracingContextFactory()
        else:
            # LangChain case
            tracing_context_factory = LangChainTracingContextFactory()

        return tracing_context_factory

    @staticmethod
    def get_context_type(config: Dict[str, Any]) -> str:
        """
        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        :return: The context type for the config
        """
        empty: Dict[str, Any] = {}
        use_config: Dict[str, Any] = config
        if use_config is None:
            use_config = empty

        # Prepare for sanity in checks below
        context_type: str = use_config.get("context_type")
        if context_type is None:
            context_type = "langchain"
        context_type = context_type.lower()

        return context_type




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import Union

import html

from neuro_san.session.mcp_service_agent_session import MCP_VERSION


class McpRequestUtil:
    """
    Utility class for generating MCP protocol requests and responses.
    """

    @classmethod
    def get_mcp_version(cls) -> str:
        """
        Get the MCP protocol version supported by this service.
        :return: MCP protocol version string.
        """
        return MCP_VERSION

    @classmethod
    def get_handshake_response(cls, request_id) -> Dict[str, Any]:
        """
        Generate a standard MCP handshake response.
        :param request_id: MCP request id;
        :return: json dictionary with handshake request in MCP format suitable for sending to a client.
        """
        return {
            "jsonrpc": "2.0",
            "id": McpRequestUtil.safe_request_id(request_id),
            "result": {
                "protocolVersion": MCP_VERSION,
                "capabilities": {
                    "logging": {},
                    "prompts": {},
                    "resources": {},
                    "tools": {
                        "listChanged": False
                    }
                },
                "serverInfo": {
                    "name": "Neuro-san-MCPServer",
                    "title": "Neuro-san MCP Server",
                    "version": "1.0.0"
                },
                "instructions": ""
            }
        }

    @staticmethod
    def safe_request_id(request_id: Union[int, str]) -> str:
        """
        Return HTML-safe representation of user request id to be sent back in MCP response.
        :param request_id: MCP request id (as received from user);
        :return: HTML-escaped request id string
        """
        # Always return a string and always HTML-escape it to avoid XSS
        # vulnerabilities in any HTML-based consumers of the MCP response.
        if isinstance(request_id, str):
            return html.escape(request_id)
        # For non-string IDs (including integers), convert to string first,
        # then escape to ensure the returned value is HTML-safe.
        return html.escape(str(request_id))

    @staticmethod
    def safe_message(msg: str) -> str:
        """
        Return HTML-safe representation of string message to be sent back in MCP response.
        :param msg: message string;
        :return: HTML-escaped message string
        """
        return html.escape(msg)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage
from langchain_core.messages import AIMessageChunk
from langchain_core.messages import BaseMessage
from langchain_core.outputs import ChatGeneration
from langchain_core.outputs import ChatGenerationChunk
from langchain_core.outputs import ChatResult
from pydantic import ConfigDict
from pydantic import Field
from tiktoken import get_encoding


class ChatMockLlm(BaseChatModel):
    """
    A custom chat model that echoes the input.

    Adapted from https://python.langchain.com/docs/how_to/custom_chat_model/
    """

    # This is required field and it is possible to have multiple test models.
    model_name: str = Field(default=None, alias="model")
    # Maybe useful for testing
    max_retries: Optional[int] = None

    # Accept both argument name and alias
    model_config = ConfigDict(populate_by_name=True)

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """Override the _generate method to implement the chat model logic.

        This can be a call to an API, a call to a local model, or any other
        implementation that generates a response to the input prompt.

        :param messages: the prompt composed of a list of messages.
        :param stop: a list of strings on which the model should stop generating.
                  If generation stops due to a stop token, the stop token itself
                  SHOULD BE INCLUDED as part of the output. This is not enforced
                  across models right now, but it's a good practice to follow since
                  it makes it much easier to parse the output of the model
                  downstream and understand why generation stopped.
        :param run_manager: A run manager with callbacks for the LLM.

        :return: chat result containing chat generation which includes ai message.
        """

        # The last message should be human message
        last_message = messages[-1]
        content = last_message.content
        input_tokens = self._num_tokens_from_string(content)
        message = AIMessage(
            content=content,
            additional_kwargs={},  # Used to add additional payload to the message
            response_metadata={  # Use for response metadata
                "model_name": self.model_name,
            },
            usage_metadata={
                "input_tokens": input_tokens,
                "output_tokens": input_tokens,
                "total_tokens": 2*input_tokens,
            },
        )

        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation])

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        """Stream the output of the model.

        This method should be implemented if the model can generate output
        in a streaming fashion.

        ***This is required for AgentExecutor.***

        Note that the _astream implementation uses run_in_executor to launch the sync _stream
        in a separate thread if _stream is implemented. Thus, it is not required to override
        the async method.

        :param messages: the prompt composed of a list of messages.
        :param stop: a list of strings on which the model should stop generating.
                  If generation stops due to a stop token, the stop token itself
                  SHOULD BE INCLUDED as part of the output. This is not enforced
                  across models right now, but it's a good practice to follow since
                  it makes it much easier to parse the output of the model
                  downstream and understand why generation stopped.
        :param run_manager: A run manager with callbacks for the LLM.
        :yields: ChatGenerationChunk objects containing the streamed model output.
        """

        # The last message should be human message
        last_message = messages[-1]
        content = last_message.content
        input_tokens = self._num_tokens_from_string(content)
        for i, content_chunk in enumerate(content):
            # This is to make input = output tokens for streaming
            if i == 0:
                output_tokens = input_tokens - len(content) + 1
            else:
                output_tokens = 1

            chunk = ChatGenerationChunk(
                message=AIMessageChunk(
                    content=content_chunk,
                    usage_metadata={
                        "input_tokens": input_tokens,
                        "output_tokens": output_tokens,
                        "total_tokens": input_tokens + output_tokens,
                    },
                )
            )
            input_tokens = 0

            if run_manager:
                # This is optional in newer versions of LangChain
                # The on_llm_new_token will be called automatically
                run_manager.on_llm_new_token(content_chunk, chunk=chunk)

            yield chunk

        # Add model name in response metadata.
        chunk = ChatGenerationChunk(
            message=AIMessageChunk(
                content="",
                response_metadata={"model_name": self.model_name},
            )
        )
        if run_manager:
            # This is optional in newer versions of LangChain
            # The on_llm_new_token will be called automatically
            run_manager.on_llm_new_token(content, chunk=chunk)

        yield chunk

    def _num_tokens_from_string(self, string: str, encoding_name: str = "o200k_base") -> int:
        """
        Returns the number of tokens in a text string using tiktoken.
        The default encoding is the same one as gpt-4o.

        :param string: Input string.
        :param encoding_name: Encoding model to use.

        :return: Number of token.
        """
        encoding = get_encoding(encoding_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens

    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model."""
        return "echoing-chat-model-basic"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters.

        This information is used by the LangChain callback system, which
        is used for tracing purposes make it possible to monitor LLMs.
        """
        return {
            # The model name allows users to specify custom token counting
            # rules in LLM monitoring application.
            "model_name": self.model_name,
        }




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from logging import getLogger
from logging import Logger

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator


class MissingNodesNetworkValidator(AbstractNetworkValidator):
    """
    AbstractNetworkValidator that looks for missing nodes that are referred to
    in an agent network.
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: A list of error messages
        """
        errors: List[str] = []

        # Validate that agent tools have corresponding nodes
        missing_nodes: Dict[str, List[str]] = self.find_missing_agent_nodes(name_to_spec)
        if missing_nodes:
            for agent, missing_tools in missing_nodes.items():
                # Format the comma-separated list of missing tools
                tools_str: str = ", ".join(f"'{tool}'" for tool in missing_tools)
                errors.append(
                    f"Agent '{agent}' references non-existent agent(s) in tools: {tools_str}"
                )

        if len(errors) > 0:
            # Only warn if there is a problem
            self.logger.warning(str(errors))

        return errors

    def find_missing_agent_nodes(self, name_to_spec: Dict[str, Any]) -> Dict[str, List[str]]:
        """
        Find agents referenced in "tools" lists that don't have corresponding nodes in the network.

        :param name_to_spec: The agent network to validate
        :return: Dictionary mapping agent names to list of tools that reference non-existent agents
                Format: {agent_name: [missing_tool1, missing_tool2, ...]}
        """
        missing_nodes: Dict[str, List[str]] = {}

        # Iterate through all agents in the network
        for agent_name, agent_data in name_to_spec.items():

            tools: List[str] = agent_data.get("tools", [])
            safe_tools: List[str] = self.remove_dictionary_tools(tools)

            # Check each tool in the agent's tools list
            for tool in safe_tools:
                # Skip URL/path tools - they're not agents and don't need nodes
                if self.is_url_or_path(tool):
                    continue

                # If tool is an agent reference but has no node in network, it's invalid
                if tool not in name_to_spec:
                    if agent_name not in missing_nodes:
                        missing_nodes[agent_name] = []
                    missing_nodes[agent_name].append(tool)

        return missing_nodes




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict


class StructureParser:
    """
    Interface that allows parsing a dictionary from within LLM response text.
    Concrete subclasses implement parsing from various formats.

    The notion of a "remainder" allows for communicating other descriptive text
    outside of what was parsed for the dictionary itself.
    """

    def __init__(self):
        """
        Constructor
        """
        self.remainder: str = None

    def parse_structure(self, content: str) -> Dict[str, Any]:
        """
        Parse the single string content for any signs of structure

        :param content: The string to parse for structure
        :return: A dictionary structure that was embedded in the content.
                Will return None if no parseable structure is detected.
        """
        raise NotImplementedError

    def get_remainder(self) -> str:
        """
        :return: Any content string that was not essential in detecting or
                 describing the structure.  The parse_structure() method must
                 be called to get anything valid out of the return value here.
                 Will return None if no parseable structure is detected.
        """
        return self.remainder




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class BedrockLlmPolicy(LlmPolicy):
    """
    LlmPolicy implementation for Bedrock.

    Bedrock does not allow for passing in async web clients.
    As a matter of fact, all of its clients are synchronous,
    which is not the best for an async service.
    """

    def create_llm(self, config: Dict[str, Any], model_name: str, client: Any) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel instance from the fully-specified llm config
        for the llm class that the implementation supports.  Chat models are usually
        per-provider, where the specific model itself is an argument to its constructor.

        :param config: The fully specified llm config
        :param model_name: The name of the model
        :param client: The web client to use (if any)
        :return: A BaseLanguageModel (can be Chat or LLM)
        """
        # Use lazy loading to prevent installing the world
        # pylint: disable=invalid-name
        ChatBedrock = self.resolver.resolve_class_in_module("ChatBedrock",
                                                            module_name="langchain_aws",
                                                            install_if_missing="langchain-aws")
        llm = ChatBedrock(
            model=model_name,
            aws_access_key_id=self.get_value_or_env(config, "aws_access_key_id", "AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=self.get_value_or_env(config, "aws_secret_access_key", "AWS_SECRET_ACCESS_KEY"),
            aws_session_token=self.get_value_or_env(config, "aws_session_token", "AWS_SESSION_TOKEN"),
            base_model_id=config.get("base_model_id"),
            beta_use_converse_api=config.get("beta_use_converse_api"),
            cache=config.get("cache"),
            config=config.get("config"),
            credentials_profile_name=config.get("credentials_profile_name"),
            custom_get_token_ids=config.get("custom_get_token_ids"),
            endpoint_url=config.get("endpoint_url"),
            guardrails=config.get("guardrails"),
            max_tokens=config.get("max_tokens"),
            metadata=config.get("metadata"),
            provider=config.get("provider"),
            rate_limiter=config.get("rate_limiter"),
            region_name=config.get("region_name"),
            stop_sequences=config.get("stop_sequences"),
            streaming=True,
            system_prompt_with_tools=config.get("system_prompt_with_tools"),
            tags=config.get("tags"),
            temperature=config.get("temperature"),

            # If omitted, this defaults to the global verbose value,
            # accessible via langchain_core.globals.get_verbose():
            # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
            #
            # However, accessing the global verbose value during concurrent initialization
            # can trigger the following warning:
            #
            # UserWarning: Importing verbose from langchain root module is no longer supported.
            # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
            # old_verbose = langchain.verbose
            #
            # To prevent this, we explicitly set verbose=False here (which matches the default
            # global verbose value) so that the warning is never triggered.
            verbose=False,
        )
        return llm

    async def delete_resources(self):
        """
        Release the run-time resources used by the model
        """
        if self.llm is None:
            return

        # Do the necessary reach-ins to successfully shut down the web client
        if self.llm.client is not None:
            # This is a boto3 client
            self.llm.client.close()

        if self.llm.bedrock_client is not None:
            # This is a boto3 client
            self.llm.bedrock_client.close()

        # Let's not do this again, shall we?
        self.llm = None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import os
import shutil

import argparse
import json

from pathlib import Path
from timedinput import timedinput

from leaf_common.config.file_of_class import FileOfClass

from neuro_san.client.agent_session_factory import AgentSessionFactory
from neuro_san.client.concierge_session_factory import ConciergeSessionFactory
from neuro_san.client.streaming_input_processor import StreamingInputProcessor
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.interfaces.concierge_session import ConciergeSession


class AgentCli:
    """
    Command line tool for communicating with a Agent service
    running in a container.
    """

    DEFAULT_INPUT: str = "DEFAULT"
    DEFAULT_PROMPT: str = "Please enter your response ('quit' to terminate):\n"
    TIMEOUT_SIGNAL: str = "<===TIMEOUT===>"

    def __init__(self):
        """
        Constructor
        """
        self.poll_timeout_seconds: float = 5.0
        self.input_timeout_seconds: float = 5000.0
        self.args = None
        self.arg_groups: Dict[str, Any] = {}

        self.session: AgentSession = None
        self.thinking_dir: str = None

    def main(self):
        """
        Main entry point for command line user interaction
        """
        self.parse_args()

        # See if we are doing a list operation
        if self.args.list or self.args.tags or self.args.tag:
            self.list()
            return

        self.open_session()
        if self.args.connectivity:
            self.connectivity()
        else:
            self.chat()

    def connectivity(self):
        """
        Perform connectivity function
        """
        empty: Dict[str, Any] = {}
        response: Dict[str, Any] = self.session.connectivity(empty)

        empty_list: List[Dict[str, Any]] = []
        for connectivity_info in response.get("connectivity_info", empty_list):
            print(f"{json.dumps(connectivity_info, indent=4, sort_keys=True)}")

        metadata = response.get("metadata")
        metadata_str: str = "<None>"
        if metadata is not None:
            metadata_str = json.dumps(metadata, indent=4, sort_keys=True)

        print(f"\nNetwork Metadata: {metadata_str}")

    def chat(self):
        """
        Perform chat function
        """

        # Get initial user input
        user_input: str = None
        if self.args.first_prompt_file is not None:
            # Incorrectly flagged as destination of Path Traversal 4
            #   Reason: first_prompt_file was previously checked with FileOfClass.check_file()
            #           which actually does the path traversal check. CheckMarx does not
            #           recognize pathlib as a valid library with which to resolve these kinds
            #           of issues.  Furthermore, this is a client command line tool that is never
            #           used inside servers which just happens to be part of a library offering.
            prompt: Path = Path(self.args.first_prompt_file)
            with prompt.open('r', encoding="utf-8") as prompt_file:
                user_input = prompt_file.read()

        sly_data: Dict[str, Any] = None
        if self.args.sly_data is not None:
            sly_data = json.loads(self.args.sly_data)
            print(f"sly_data is {sly_data}")

        # Note: If nothing is specified the server assumes the chat_filter_type
        #       should be "MINIMAL", however for this client which is aimed at
        #       developers, we specifically want a default MAXIMAL client to
        #       show all the bells and whistles of the output that a typical
        #       end user will not care about and not appreciate the extra
        #       data charges on their cell phone.
        chat_filter: Dict[str, Any] = {
            "chat_filter_type": "MAXIMAL"
        }
        if self.args.chat_filter is not None and \
                not bool(self.args.chat_filter):
            chat_filter["chat_filter_type"] = "MINIMAL"

        message = f"""
The agent "{self.args.agent}" is not implemented on the server.

Some suggestions:
1. Did you misspell the agent name on the command line?
2. Is there a key for the agent name in the server manifest.hocon file?
3. Is the value for the agent name key in the server manifest.hocon file set to true?
4. Servers will skip manifest entries that have errors. They will also print out which
   agents they are actually serving.  Check your server output for each of these.
5. Is the server itself actually running?
"""

        empty: Dict[str, Any] = {}
        response: Dict[str, Any] = self.session.function(empty)
        if response is None:
            raise ValueError(message)

        function: Dict[str, Any] = response.get("function", empty)
        initial_prompt: str = function.get("description")
        print(f"\n{initial_prompt}\n")

        if not self.args.no_thinking_file:
            print("To see the thinking involved with the agent:\n")
            if not self.args.thinking_dir:
                print(f"    tail -f {self.args.thinking_file}\n")
            else:
                print(f"    See any one of the files in {self.thinking_dir} for agent network chat details.\n")

        self.loop_until_done(user_input, sly_data, chat_filter)

    def loop_until_done(self, user_input: str, sly_data: Dict[str, Any],
                        chat_filter: Dict[str, Any]):
        """
        Loop trading off between user input and agent response until done.
        :param user_input: The initial user input (if any)
        :param sly_data: The initial sly data dictionary (if any)
        :param chat_filter: The chat filter dictionary to use
        """

        state: Dict[str, Any] = {
            "last_chat_response": None,
            "prompt": self.DEFAULT_PROMPT,
            "timeout": self.input_timeout_seconds,
            "num_input": 0,
            "user_input": user_input,
            "sly_data": sly_data,
            "chat_filter": chat_filter,
        }

        input_processor = StreamingInputProcessor(self.DEFAULT_INPUT,
                                                  self.args.thinking_file,
                                                  self.session,
                                                  self.thinking_dir)

        while not self.is_done(state):

            prompt = state.get("prompt")
            timeout = state.get("timeout")
            user_input = state.get("user_input")
            if user_input is None:
                if prompt is not None and len(prompt) > 0:
                    user_input = timedinput(prompt, timeout=timeout,
                                            default=self.TIMEOUT_SIGNAL)
                    if user_input == self.TIMEOUT_SIGNAL:
                        # Timeout occurred - nobody is talking to us.
                        print(f"\nNo input received after {timeout} seconds. Exiting.\n")
                        break
                else:
                    user_input = None

            if user_input == "quit":
                break

            print(f"Sending user_input {user_input}")
            state["user_input"] = user_input
            state = input_processor.process_once(state)

            print(f"\nResponse from {state.get('origin_str')}:")
            print(f"{state.get('last_chat_response')}")
            if state.get("sly_data") is not None:
                pretty_sly: str = json.dumps(state.get('sly_data'), indent=4, sort_keys=True)
                print(f"Returned sly_data is: {pretty_sly}")

            if self.args.tokens:
                print("Token Accounting:")
                print(f"{json.dumps(state.get('token_accounting'), indent=4, sort_keys=True)}")

            if self.args.response_output_file is not None:
                output_path: Path = Path(self.args.response_output_file)
                with output_path.open('w', encoding="utf-8") as output_file:
                    output_file.write(state["last_chat_response"])
                    output_file.write("\n")

    def parse_args(self):
        """
        Parse command line arguments into member variables
        """

        arg_parser = argparse.ArgumentParser()

        self.add_args(arg_parser)

        # Incorrectly flagged as source of Path Traversal 1, 2, 4, 5, 6
        # See destination in file_of_class.py for exception explanation.
        # Incorrectly flagged as source of Trust Boundary Violation 1, 2
        # See destination in agent_session_factory.py for exception explanation.
        self.args = arg_parser.parse_args()

        # Check some arguments to prevent PathTraversal scans lighting up.
        # Since this is a command-line tool not intended to be used inside a
        # Dockerfile service, we don't really care where these files come from.
        # Check anyway to give warm fuzzies to scans.
        if not self.args.no_thinking_file:
            self.args.thinking_file = FileOfClass.check_file(self.args.thinking_file, "/")
        self.args.first_prompt_file = FileOfClass.check_file(self.args.first_prompt_file, "/")

    def add_args(self, arg_parser: argparse.ArgumentParser):
        """
        Adds arguments.  Allows subclasses a chance to add their own.
        :param arg_parser: The argparse.ArgumentParser to add.
        """
        # What agent are we talking to?
        arg_parser.add_argument("--agent", type=str, default="esp_decision_assistant",
                                help="Name of the agent to talk to")

        # What agents can we connect to?
        group = arg_parser.add_argument_group(title="Agent Listings",
                                              description="What agents can we connect to?")
        group.add_argument("--list", default=False, dest="list", action="store_true",
                           help="List all available agents")
        group.add_argument("--tag", type=str,
                           help="List all agents marked with the given tag.")
        group.add_argument("--tags", default=False, dest="tags", action="store_true",
                           help="List all tags associated with agents")

        # How will we connect to neuro-san?
        group = arg_parser.add_argument_group(title="Session Type",
                                              description="How will we connect to neuro-san?")
        group.add_argument("--connection", default="direct", type=str,
                           choices=["direct", "http", "https"],
                           help="""
The type of connection to initiate. Choices are to connect to:
    "http"      - an agent service via HTTP. Needs host and port.
    "https"     - an agent service via secure HTTP. Needs host and port.
    "mcp"       - an agent service via MCP protocol. Needs host and port.
    "direct"    - a session via library. (The default).
All choices require an agent name.
""")
        group.add_argument("--direct", dest="connection", action="store_const", const="direct",
                           help="Use a direct/library call for the chat")
        group.add_argument("--http", dest="connection", action="store_const", const="http",
                           help="Use a HTTP service connection")
        group.add_argument("--https", dest="connection", action="store_const", const="https",
                           help="Use a secure HTTP service connection. "
                                "Requires your agent server to be set up with certificates that are well known. "
                                "This is not something that our basic server setup supports out-of-the-box.")
        group.add_argument("--mcp", dest="connection", action="store_const", const="mcp",
                           help="Use an MCP protocol service connection.")
        group.add_argument("--timeout", dest="timeout", type=float,
                           help="Timeout in seconds before giving up on connecting to a server. "
                                "By default this is None, implying we will try forever")
        self.arg_groups[group.title] = group

        # How will we connect to a server?
        group = arg_parser.add_argument_group(title="Service Connection",
                                              description="How will we connect to a server?")
        group.add_argument("--local", default=True, action="store_true",
                           help="If True (the default), assume we are running against locally running service")
        group.add_argument("--host", type=str, default=None,
                           help="hostname setting if not running locally")
        group.add_argument("--port", type=int, default=AgentSession.DEFAULT_PORT,
                           help="TCP/IP port to run the Agent http service on")
        group.add_argument("--user_id", default=os.environ.get("USER"), type=str,
                           help="'user_id' metadata to send to a server for logging. Defaults to ${USER}.")
        self.arg_groups[group.title] = group

        # How can we get input to the chat client without typing it in?
        group = arg_parser.add_argument_group(title="Input Control",
                                              description="How do we get input without typing it it?")
        group.add_argument("--sly_data", type=str,
                           help="JSON string containing data that is out-of-band to the chat stream, "
                                "but is still essential to agent function")
        group.add_argument("--first_prompt_file", type=str,
                           help="File that captures the first response to the input prompt")
        group.add_argument("--max_input", type=int, default=1000000,
                           help="Maximum rounds of input to go before exiting")
        group.add_argument("--one_shot", dest="max_input", action="store_const", const=1,
                           help="Send one round of input, then exit")
        self.arg_groups[group.title] = group

        # How do we handle calls to external agents?
        group = arg_parser.add_argument_group(title="Local External Agents",
                                              description="How do handle calls to local /external agents?")
        group.add_argument("--local_externals_direct", default=False, action="store_true",
                           help="""
Have external tools that can be found in the local agent manifest use a
direct connection instead of requiring a service to be stood up.
                           """)
        group.add_argument("--local_externals_service", dest="local_externals_direct", action="store_false",
                           help="""
Have external tools that can be found in the local agent manifest use a service connection. (The default)
                           """)
        self.arg_groups[group.title] = group

        # How do we receive messages?
        group = arg_parser.add_argument_group(title="Message Filtering",
                                              description="What kind of messages will we receive?")
        group.add_argument("--maximal", default=True, dest="chat_filter", action="store_true",
                           help="Allow all messages to come from the server")
        group.add_argument("--minimal", dest="chat_filter", action="store_false",
                           help="Allow only the bare minimum of messages to come from the server")
        self.arg_groups[group.title] = group

        # How are we capturing output?
        group = arg_parser.add_argument_group(title="Output Capture",
                                              description="How will we capture output?")
        group.add_argument("--thinking_file", type=str, default="/tmp/agent_thinking.txt",
                           help="File that captures agent thinking. "
                                "This is a separate text stream from the user/assistant chat")
        group.add_argument("--no_thinking_file", default=False, action="store_true",
                           help="If set, blocks creation of thinking file and any output to it.")
        group.add_argument("--thinking_dir", default=True, action="store_true",
                           help="Use the basis of the thinking_file as a directory to capture "
                                "internal agent chatter in separate files. "
                                "This is a separate text stream from the user/assistant chat. "
                                "Only available when streaming (which is the default).")
        group.add_argument("--response_output_file", type=str,
                           help="File that captures the response to the user input")
        self.arg_groups[group.title] = group

        # How are we capturing output?
        group = arg_parser.add_argument_group(title="Other Modes of Operation")
        group.add_argument("--connectivity", default=False, dest="connectivity", action="store_true",
                           help="List the connectivity of agent network nodes")
        group.add_argument("--tokens", default=False, dest="tokens", action="store_true",
                           help="Print token accounting info with every response")
        self.arg_groups[group.title] = group

    def open_session(self):
        """
        Opens a session based on the parsed command line arguments
        """
        hostname = "localhost"
        if self.args.host is not None or not self.args.local:
            hostname = self.args.host

        # Open a session with the factory
        factory: AgentSessionFactory = self.get_agent_session_factory()
        metadata: Dict[str, str] = {
            "user_id": self.args.user_id
        }
        self.session = factory.create_session(self.args.connection, self.args.agent,
                                              hostname, self.args.port, self.args.local_externals_direct,
                                              metadata, self.args.timeout)

        if self.args.no_thinking_file:
            # Don't create a thinking file or directory at all. This will disable any output to a thinking file.
            self.args.thinking_file = None
            self.thinking_dir = None
            self.args.thinking_dir = False
            return

        # Clear out the previous thinking file/dir contents
        #
        # Incorrectly flagged as destination of Path Traversal 5
        #   Reason: thinking_file was previously checked with FileOfClass.check_file()
        #           which actually does the path traversal check. CheckMarx does not
        #           recognize pathlib as a valid library with which to resolve these kinds
        #           of issues.  Furthermore, this is a client command line tool that is never
        #           used inside servers which just happens to be part of a library offering.
        if not self.args.thinking_dir:
            with open(self.args.thinking_file, "w", encoding="utf-8") as thinking:
                thinking.write("\n")
        else:
            # Use the stem of the thinking file (i.e. no ".txt" extension) as the
            # basis for the thinking directory
            self.thinking_dir, extension = os.path.splitext(self.args.thinking_file)
            _ = extension

            # Remove any contents that might be there already.
            # Writing over existing dir will just confuse output.
            if os.path.exists(self.thinking_dir):
                shutil.rmtree(self.thinking_dir)
            # Create the directory anew
            os.makedirs(self.thinking_dir)

    def get_agent_session_factory(self) -> AgentSessionFactory:
        """
        This allows subclasses to add different kinds of connections.

        :return: An AgentSessionFactory instance that will allow creation of the
                 session with the agent network.
        """
        return AgentSessionFactory()

    def is_done(self, state: Dict[str, Any]) -> bool:
        """
        :param state: The state dictionary
        :return: True if the values in the state are considered to be sufficient
                for terminating further conversation. False otherwise.
        """

        if state.get("user_input") == "quit":
            return True

        if state.get("num_input") >= self.args.max_input:
            return True

        return False

    def list(self):
        """
        List available agents via a concierge session
        """
        hostname = "localhost"
        if self.args.host is not None or not self.args.local:
            hostname = self.args.host

        metadata: Dict[str, str] = {
            "user_id": self.args.user_id
        }

        concierge_session: ConciergeSession = ConciergeSessionFactory().create_session(
            self.args.connection,
            hostname=hostname,
            port=self.args.port,
            metadata=metadata,
            connect_timeout_in_seconds=self.args.timeout)

        request_dict: Dict[str, Any] = {}
        response_dict: Dict[str, Any] = concierge_session.list(request_dict)

        empty_list: List[Dict[str, Any]] = []
        if self.args.tags:
            tags = set()
            for agent_info in response_dict.get("agents", empty_list):
                agent_tags: List[str] = agent_info.get("tags", empty_list)
                tags.update(agent_tags)
            print(f"Available tags:\n{json.dumps(list(tags), indent=4, sort_keys=True)}")

        elif self.args.tag:
            print(f"Available agents for tag {self.args.tag}:\n")
            for agent_info in response_dict.get("agents", empty_list):
                agent_tags: List[str] = agent_info.get("tags", empty_list)
                if self.args.tag in agent_tags:
                    print(json.dumps(agent_info, indent=4, sort_keys=True))

        else:
            print(f"Available agents:\n{json.dumps(response_dict, indent=4, sort_keys=True)}")


if __name__ == '__main__':
    AgentCli().main()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict

from langchain_core.runnables.base import Runnable
from langchain_core.runnables.passthrough import RunnablePassthrough

from neuro_san.internals.interfaces.run_target import RunTarget
from neuro_san.internals.run_context.langchain.tracing.neuro_san_runnable import NeuroSanRunnable


class LangChainTracingContext(RunTarget):
    """
    RunTarget interface for a TracingContext in langchain.
    """

    def __init__(self, run_target: RunTarget, config: Dict[str, Any]):
        """
        Constructor

        :param run_target: The RunTarget instance to be traced
        :param config: The configuration for the tracing context
        """
        self.run_target: RunTarget = run_target
        self.config: Dict[str, Any] = config

    async def run_it(self, inputs: Any) -> Any:
        """
        Entry point method for the run.

        :param inputs: A list of inputs from the user.
        :return: The outputs of the run.
        """
        runnable = NeuroSanRunnable(run_target=self.run_target, **self.config)
        runnable_config: Dict[str, Any] = runnable.prepare_runnable_config(use_run_name=True)

        chain: Runnable = RunnablePassthrough() | runnable

        await chain.ainvoke(input=inputs, config=runnable_config)

        return inputs




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.filters.chat_context_message_filter import ChatContextMessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class ChatContextMessageProcessor(MessageProcessor):
    """
    Implementation of the MessageProcessor that looks for the chat_context
    in the stream which is used to continue the conversation.
    """

    def __init__(self):
        """
        Constructor
        """
        self.chat_context: Dict[str, Any] = {}
        self.sly_data: Dict[str, Any] = None
        self.filter = ChatContextMessageFilter()

    def get_chat_context(self) -> Dict[str, Any]:
        """
        :return: The chat_context discovered from the agent session interaction
                Empty dictionaries or None values simply start a new conversation.
        """
        return self.chat_context

    def get_sly_data(self) -> Dict[str, Any]:
        """
        :return: The sly_data discovered from the agent session interaction
        """
        return self.sly_data

    def reset(self):
        """
        Resets any previously accumulated state
        """
        self.chat_context = {}
        self.sly_data = None

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        if not self.filter.allow_message(chat_message_dict, message_type):
            # Filter says no
            return

        # Normally the very last message holds the chat_context.
        # Keep accumulating until it comes past, as long as there is something.
        self.chat_context = chat_message_dict.get("chat_context", self.chat_context)
        self.sly_data = chat_message_dict.get("sly_data", self.sly_data)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.core.langchain_run_context import LangChainRunContext


class RunContextFactory:
    """
    Creates the correct kind of RunContext
    """

    @staticmethod
    def create_run_context(parent_run_context: RunContext,
                           tool_caller: ToolCaller,
                           invocation_context: InvocationContext = None,
                           chat_context: Dict[str, Any] = None,
                           config: Dict[str, Any] = None) \
            -> RunContext:
        """
        Creates an appropriate RunContext

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param tool_caller: The ToolCaller whose lifespan matches that
                            of the newly created RunContext
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        """

        # Initialize return value
        run_context: RunContext = None

        empty: Dict[str, Any] = {}
        use_config: Dict[str, Any] = config
        if use_config is None:
            use_config = empty

        # Get some fields from the config with reasonable defaults
        default_llm_config: Dict[str, Any] = {
            "model_name": "gpt-5.2",
            "verbose": False
        }
        default_llm_config = use_config.get("llm_config") or default_llm_config

        # Prepare for sanity in checks below
        context_type: str = MasterLlmFactory.get_context_type(use_config)

        use_invocation_context: InvocationContext = invocation_context
        if use_invocation_context is None and parent_run_context is not None:
            use_invocation_context = parent_run_context.get_invocation_context()

        if context_type.startswith("openai"):
            raise ValueError("OpenAI Assistants implementation is no longer supported by OpenAI.")

        if context_type.startswith("langchain"):
            run_context = LangChainRunContext(default_llm_config, parent_run_context,
                                              tool_caller, use_invocation_context,
                                              chat_context, use_config.get("middleware_config"))
        else:
            # Default case
            run_context = LangChainRunContext(default_llm_config, parent_run_context,
                                              tool_caller, use_invocation_context,
                                              chat_context, use_config.get("middleware_config"))

        return run_context




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from pathlib import Path

from leaf_common.config.config_filter_chain import ConfigFilterChain

from neuro_san.internals.interfaces.agent_name_mapper import AgentNameMapper
from neuro_san.internals.graph.persistence.agent_filetree_mapper import AgentFileTreeMapper
from neuro_san.internals.graph.persistence.agent_standalone_mapper import AgentStandaloneMapper
from neuro_san.internals.graph.filters.defaults_config_filter import DefaultsConfigFilter
from neuro_san.internals.graph.filters.dictionary_common_defs_config_filter \
    import DictionaryCommonDefsConfigFilter
from neuro_san.internals.graph.filters.name_correction_config_filter import NameCorrectionConfigFilter
from neuro_san.internals.graph.filters.string_common_defs_config_filter import StringCommonDefsConfigFilter
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.persistence.abstract_async_config_restorer import AbstractAsyncConfigRestorer


class AgentNetworkRestorer(AbstractAsyncConfigRestorer):
    """
    Implementation of the AbstractAsyncConfigRestorer interface to read in an AgentNetwork
    instance given a JSON file name.
    """

    def __init__(self, registry_dir: str = None, agent_mapper: AgentNameMapper = None):
        """
        Constructor

        :param registry_dir: The directory under which file_references
                    for registry files are allowed to be found.
                    If None, there are no limits, but paths must be absolute
        :param agent_mapper: optional AgentNameMapper;
            if None, default will be used:
                if registry_dir is None, AgentStandaloneMapper instance will be used;
                otherwise, we use AgentFileTreeMapper.
        """
        super().__init__(file_purpose="agent network", must_exist=True)

        self.registry_dir: str = registry_dir
        self.agent_mapper = agent_mapper
        if not self.agent_mapper:
            if self.registry_dir is not None:
                self.agent_mapper = AgentFileTreeMapper()
            else:
                self.agent_mapper = AgentStandaloneMapper()

    def get_file_path(self, file_reference: str = None) -> str:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: an string of the file path to use
        """
        if file_reference is None or len(file_reference) == 0:
            raise ValueError(f"file_reference {file_reference} cannot be None or empty string")

        use_file: str = file_reference
        if self.registry_dir is not None:
            # This should be OS-agnostic operation, producing a valid local file path
            use_file = str(Path(self.registry_dir) / file_reference)

        return use_file

    def restore(self, file_reference: str = None) -> AgentNetwork:
        """
        Synchronous restore from the given file reference.
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: an AgentNetwork.
        """
        config: Dict[str, Any] = super().restore(file_reference)
        return self.create_network(config, file_reference)

    async def async_restore(self, file_reference: str = None) -> AgentNetwork:
        """
        Asynchronous restore from the given file reference.
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: an AgentNetwork.
        """
        config: Dict[str, Any] = await super().async_restore(file_reference)
        return self.create_network(config, file_reference)

    def create_network(self, config: Dict[str, Any], file_reference: str) -> AgentNetwork:
        """
        :param config: agent configuration dictionary, built or parsed from external sources
        :param file_reference: The file reference used for restoring
        :return: AgentNetwork instance for an agent.
        """
        if config is None:
            return None

        # Inside here is incorrectly flagged as destination of Path Traversal 7
        #   Reason: The lines above ensure that the path of registry_dir is within
        #           this source base. CheckMarx does not recognize
        #           the calls to Pathlib/__file__ as a valid means to resolve
        #           these kinds of issues.
        name: str = self.agent_mapper.filepath_to_agent_network_name(file_reference)

        # Now create the AgentNetwork
        agent_network = AgentNetwork(config, name)
        return agent_network

    def filter_config(self, basis_config: Dict[str, Any], file_path: str = None) -> Dict[str, Any]:
        """
        :param basis_config: agent configuration dictionary, built or parsed from external sources
        :return: An agent network dictionary that has gone through the standard filter chain.
        """
        # Perform a filter chain on the config that was read in
        filter_chain = ConfigFilterChain()
        filter_chain.register(DictionaryCommonDefsConfigFilter())
        filter_chain.register(StringCommonDefsConfigFilter())
        filter_chain.register(DefaultsConfigFilter())
        filter_chain.register(NameCorrectionConfigFilter())

        config: Dict[str, Any] = filter_chain.filter_config(basis_config)
        return config




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List
from typing import Set

import copy
import re
import jsonschema

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator


class ToolRequestValidator(DictionaryValidator):
    """
    Class implementing MCP tool call request validation against tool call schema.
    """
    def __init__(self, service_schema: Dict[str, Any]):
        """
        Initialize the tool request validator.
        :param service_schema: The OpenAPI schema dictionary for the neuro-san service API.
        """
        self.tool_request_method = "ChatRequest"
        self.required_property = "user_message"
        # Generate the validation schema for tool call requests
        # from the overall neuro-san service schema:
        self.request_schema = self._extract_sub_schema(
            service_schema,
            self.tool_request_method,
            self.required_property)

    def validate(self, candidate: Dict[str, Any]) -> List[str]:
        """
        Validate the dictionary data of an incoming tool call request against MCP protocol schema.
        :param candidate: The request dictionary to validate
        :return: A list of error messages, if any
        """
        try:
            jsonschema.validate(instance=candidate, schema=self.request_schema)
        except jsonschema.exceptions.ValidationError:
            # We don't return detailed validation errors to a client,
            # since they tend to be very long and complex.
            return [f"Request validation FAILED for tool call request: {candidate}"]
        except Exception as exc:  # pylint: disable=broad-exception-caught
            return [f"Validation exception: {str(exc)}"]
        return None

    def get_request_schema(self) -> Dict[str, Any]:
        """
        :return: The tool request validation schema
        """
        return self.request_schema

    def _get_schema_defs(self, schema: Dict[str, Any]) -> Set[str]:
        """
        Return the set of referenced schema names found in json schema via $ref.

        Recognizes local refs of the form:
          - "#/components/schemas/Name"
        """
        _ref_defs_re = re.compile(r"^#/components/schemas/([^/]+)$")
        out: Set[str] = set()

        def visit(node: Any):
            if isinstance(node, dict):
                # Collect $ref if present
                ref = node.get("$ref")
                if isinstance(ref, str):
                    m_ref = _ref_defs_re.match(ref)
                    if m_ref:
                        out.add(m_ref.group(1))
                # Recurse through all values (including keys like allOf/anyOf/etc.)
                for v in node.values():
                    visit(v)
            elif isinstance(node, list):
                for item in node:
                    visit(item)
            # primitives: nothing to do

        visit(schema)
        return out

    def _get_next(self, items: Dict[str, bool]) -> str:
        """
        Returns the next unprocessed item in the provided dictionary.
        Iterates through the given dictionary to find the first key whose value
        is `False`, indicating it has not yet been processed.
        :param items: A dictionary where keys are items being processed
                      and values are booleans indicating if the item has been processed.
        :return: The key of the first unprocessed item, or `None` if all items are done.
        """
        for item, processed in items.items():
            if not processed:
                return item
        return None

    def _get_all_defs(self, root_item: str, schemas: Dict[str, Any]) -> Set[str]:
        """
        :param root_item: The root item to start processing from
        :param schemas: The json schemas dictionary to process
        :return: A set of all item names which were recursively referenced by the given root item
        """
        items: Dict[str, bool] = {root_item: False}
        processing: bool = True
        while processing:
            next_item = self._get_next(items)
            if next_item:
                items[next_item] = True  # Mark as being processed
                next_schema: Dict[str, Any] = schemas.get(next_item)
                if not next_schema:
                    raise ValueError(f"No schema found for {next_item}")
                referenced_defs = self._get_schema_defs(next_schema)
                for ref in referenced_defs:
                    if ref not in items:
                        items[ref] = False
            else:
                processing = False
        return set(items)

    def _extract_sub_schema(self, schema: Dict[str, Any], root_item: str, required_property: str) -> Dict[str, Any]:
        """
        Extracts a sub-schema from the given OpenAPI schema, starting from the given root item.
        """
        result: Dict[str, Any] = {}
        # Get all the schema definitions:
        schemas: Dict[str, Any] = schema.get("components", {}).get("schemas", {})
        if not schemas:
            raise ValueError("No components found in schema")
        item_schema: Dict[str, Any] = schemas.get(root_item, {})
        if not item_schema:
            raise ValueError(f"No schema found for {root_item}")
        result["components"] = {"schemas": copy.deepcopy(schemas)}
        referenced_defs = self._get_all_defs(root_item, schemas)
        # To reduce the overall size of the resulting schema,
        # remove all schemas that are not referenced by the root item.
        for one_schema in schemas:
            if one_schema not in referenced_defs:
                del result["components"]["schemas"][one_schema]
        # Add required properties to the root schema
        root_schema = result["components"]["schemas"][root_item]
        root_schema["required"] = [required_property]
        # Make tool call schema strict - no additional properties allowed
        root_schema["additionalProperties"] = False
        result["$ref"] = f"#/components/schemas/{root_item}"
        return result




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import uuid

from neuro_san.internals.run_context.interfaces.tool_call import ToolCall


class LangChainToolCall(ToolCall):
    """
    A LangChain implementation of a ToolCall

    For the uninitiated: A "ToolCall" in langchain/openai parlance is a *request*
    that a tool be called with certain structured function arguments.
    """

    def __init__(self, tool_name: str, args: Any, run_id: str, invocation: str = None):
        """
        Constructor

        :param tool_name: The name of the tool to be called
        :param args: The arguments the tool is requested to be called with
                So far we've only seen this as Dict[str, Any], but the langchain
                typing is Any, so we stick with that.
        :param run_id: The string id of the parent run so that the tool's
                    ids can be associated with that.
        :param invocation: String describing how the tool wants to be invoked.
                            Can be: "chatbot" - implies waiting for an answer.
                                    "event" - implies no answer needed
                                    None - implies chatbot
        """
        self.tool_name: str = tool_name
        self.args = args
        self.id: str = f"tool_call_{run_id}_{uuid.uuid4()}"
        self.invocation: str = invocation
        if invocation is None:
            self.invocation = "chatbot"

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.id

    def get_function_arguments(self) -> Dict[str, Any]:
        """
        :return: Returns a dictionary of the function arguments for the tool call
        """
        return self.args

    def get_function_name(self) -> str:
        """
        :return: Returns the string name of the tool
        """
        return self.tool_name

    def get_invocation(self) -> str:
        """
        :return: String describing how the tool wants to be invoked.
                            Can be: "chatbot" - implies waiting for an answer.
                                    "event" - implies no answer needed
                                    None - implies chatbot
        """
        return self.invocation




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from neuro_san.internals.filters.answer_message_filter import AnswerMessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.internals.parsers.structure.first_available_structure_parser \
    import FirstAvailableStructureParser
from neuro_san.message_processing.message_processor import MessageProcessor


class AnswerMessageProcessor(MessageProcessor):
    """
    Implementation of the MessageProcessor that looks for "the answer"
    of the chat session.
    """

    def __init__(self, structure_formats: Union[str, List[str]] = None):
        """
        Constructor

        :param structure_formats: Optional string or list of strings telling us to look for
                    specific formats within the text of the final answer to separate out
                    in a common way so that clients do not have to reinvent this wheel over
                    and over again.

                    Valid values are:
                        "json" - look for JSON in the message content as structure to report.

                    By default this is None, implying that such parsing is bypassed.
        """
        self.answer: str = None
        self.structure: Dict[str, Any] = None
        self.answer_origin: List[Dict[str, Any]] = None
        self.filter: AnswerMessageFilter = AnswerMessageFilter()

        # Only deal with non-empy lists of strings internally
        self.structure_formats: List[str] = structure_formats
        if self.structure_formats is not None:
            if isinstance(self.structure_formats, str):
                self.structure_formats = [self.structure_formats]
        else:
            self.structure_formats = []

        if not isinstance(self.structure_formats, List):
            raise ValueError(f"Value '{structure_formats}' must be a string, a list of strings, or None")

    def get_answer(self) -> str:
        """
        :return: The final answer from the agent session interaction
        """
        return self.answer

    def get_answer_origin(self) -> List[Dict[str, Any]]:
        """
        :return: The origin of the final answer from the agent session interaction
        """
        return self.answer_origin

    def get_structure(self) -> Dict[str, Any]:
        """
        :return: Any dictionary structure that was contained within the final answer
                 from the agent session interaction, if such a specific breakout was desired.
        """
        return self.structure

    def reset(self):
        """
        Resets any previously accumulated state
        """
        self.answer = None
        self.structure = None
        self.answer_origin = None

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        if not self.filter.allow_message(chat_message_dict, message_type):
            # Does not pass the criteria for a message holding a final answer
            return

        origin: List[Dict[str, Any]] = chat_message_dict.get("origin")
        text: str = chat_message_dict.get("text")
        structure: Dict[str, Any] = chat_message_dict.get("structure")

        # Record what we got.
        # We might get another as we go along, but the last message in the stream
        # meeting the criteria above is our final answer.
        if text is not None:
            self.answer = text
        if origin is not None:
            self.answer_origin = origin
        if structure is not None:
            self.structure = structure

        if structure is None and text is not None:
            # Parse structure from the first available format in the answer content
            structure_parser = FirstAvailableStructureParser(self.structure_formats)
            self.structure = structure_parser.parse_structure(text)
            if self.structure is not None:
                self.answer = structure_parser.get_remainder()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from contextlib import suppress

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class AnthropicLlmPolicy(LlmPolicy):
    """
    Implementation of the LlmPolicy for Anthtropic chat models.

    Anthropic chat models do not allow for passing in an externally managed
    async web client.
    """

    def create_llm(self, config: Dict[str, Any], model_name: str, client: Any) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel instance from the fully-specified llm config
        for the llm class that the implementation supports.  Chat models are usually
        per-provider, where the specific model itself is an argument to its constructor.

        :param config: The fully specified llm config
        :param model_name: The name of the model
        :param client: The web client to use (if any)
        :return: A BaseLanguageModel (can be Chat or LLM)
        """
        # Use lazy loading to prevent installing the world
        # pylint: disable=invalid-name
        ChatAnthropic = self.resolver.resolve_class_in_module("ChatAnthropic",
                                                              module_name="langchain_anthropic.chat_models",
                                                              install_if_missing="langchain-anthropic")

        llm = ChatAnthropic(
            model_name=model_name,
            max_tokens=config.get("max_tokens"),  # This is always for output
            temperature=config.get("temperature"),
            top_k=config.get("top_k"),
            top_p=config.get("top_p"),
            default_request_timeout=config.get("default_request_timeout"),
            max_retries=config.get("max_retries"),
            stop_sequences=config.get("stop_sequences"),
            anthropic_api_url=self.get_value_or_env(config, "anthropic_api_url",
                                                    "ANTHROPIC_API_URL"),
            anthropic_api_key=self.get_value_or_env(config, "anthropic_api_key",
                                                    "ANTHROPIC_API_KEY"),
            default_headers=config.get("default_headers"),
            betas=config.get("betas"),
            streaming=True,  # streaming is always on. Without it token counting will not work.
            # Set stream_usage to True in order to get token counting chunks.
            stream_usage=True,
            thinking=config.get("thinking"),
            mcp_servers=config.get("mcp_servers"),
            context_management=config.get("context_management"),
            # If omitted, this defaults to the global verbose value,
            # accessible via langchain_core.globals.get_verbose():
            # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
            #
            # However, accessing the global verbose value during concurrent initialization
            # can trigger the following warning:
            #
            # UserWarning: Importing verbose from langchain root module is no longer supported.
            # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
            # old_verbose = langchain.verbose
            #
            # To prevent this, we explicitly set verbose=False here (which matches the default
            # global verbose value) so that the warning is never triggered.
            verbose=False,
        )
        return llm

    async def delete_resources(self):
        """
        Release the run-time resources used by the model
        """
        if self.llm is None:
            return

        # Do the necessary reach-ins to successfully shut down the web client

        # This is really an anthropic.AsyncClient, but we don't really want to do the Resolver here.
        # Note we don't want to do this in the constructor, as AnthropicChat lazily
        # creates these as needed via a cached_property that needs to be done in its own time
        # via Anthropic infrastructure.  By the time we get here, it's already been created.
        anthropic_async_client: Any = self.llm._async_client     # pylint:disable=protected-access

        if anthropic_async_client is not None:
            with suppress(Exception):
                await anthropic_async_client.aclose()

        # Let's not do this again, shall we?
        self.llm = None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List

import jsonschema

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator


class McpRequestValidator(DictionaryValidator):
    """
    Class implementing MCP request validation against MCP protocol schema.
    """
    def __init__(self, validation_schema: Dict[str, Any]):
        self.validation_schema = validation_schema

    def validate(self, candidate: Dict[str, Any]) -> List[str]:
        """
        Validate the dictionary data of incoming MCP request against MCP protocol schema.
        :param candidate: The request dictionary to validate
        :return: A list of error messages, if any
        """
        try:
            jsonschema.validate(instance=candidate, schema=self.validation_schema)
        except jsonschema.exceptions.ValidationError:
            # We don't return detailed validation errors to the client,
            # since they tend to be very long and complex.
            return [f"Request validation FAILED for MCP request: {candidate}"]
        except Exception as exc:  # pylint: disable=broad-exception-caught
            return [f"Validation exception: {str(exc)}"]
        return None
